import io
import enum

class PTRFLAGS(enum.IntFlag):
	HWHEEL = 0x0400 #The event is a horizontal mouse wheel rotation. The only valid flags in a horizontal wheel rotation event are PTRFLAGS_WHEEL_NEGATIVE and the WheelRotationMask; all other pointer flags are ignored. This flag MUST NOT be sent to a server that does not indicate support for horizontal mouse wheel events in the Input Capability Set (section 2.2.7.1.6).
	WHEEL = 0x0200 #The event is a vertical mouse wheel rotation. The only valid flags in a vertical wheel rotation event are PTRFLAGS_WHEEL_NEGATIVE and the WheelRotationMask; all other pointer flags are ignored.
	WHEEL_NEGATIVE = 0x0100 #The wheel rotation value (contained in the WheelRotationMask bit field) is negative and MUST be sign-extended before injection at the server.
	WheelRotationMask = 0x01FF #The bit field describing the number of rotation units the mouse wheel was rotated. The value is negative if the PTRFLAGS_WHEEL_NEGATIVE flag is set.
	MOVE = 0x0800 #    Indicates that the mouse position MUST be updated to the location specified by the xPos and yPos fields.
	DOWN = 0x8000 #    Indicates that a click event has occurred at the position specified by the xPos and yPos fields. The button flags indicate which button has been clicked and at least one of these flags MUST be set.
	BUTTON1 = 0x1000 #    Mouse button 1 (left button) was clicked or released. If the PTRFLAGS_DOWN flag is set, then the button was clicked, otherwise it was released.
	BUTTON2 = 0x2000 #    Mouse button 2 (right button) was clicked or released. If the PTRFLAGS_DOWN flag is set, then the button was clicked, otherwise it was released.
	BUTTON3 = 0x4000 #    Mouse button 3 (middle button or wheel) was clicked or released. If the PTRFLAGS_DOWN flag is set, then the button was clicked, otherwise it was released.

class TS_POINTER_EVENT:
	def __init__(self):
		self.pointerFlags:PTRFLAGS = None #only release flag is allowed here
		self.xPos:int = None #maybe string? #The Unicode character input code.
		self.yPos:int = None

	def to_bytes(self):
		t = self.pointerFlags.to_bytes(2, byteorder='little', signed=False)
		t += self.xPos.to_bytes(2, byteorder='little', signed=False)
		t += self.yPos.to_bytes(2, byteorder='little', signed=False)
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return TS_POINTER_EVENT.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = TS_POINTER_EVENT()
		msg.pointerFlags = PTRFLAGS(int.from_bytes(buff.read(2), byteorder='little', signed=False))
		msg.xPos = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		msg.yPos = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		return msg

	def __repr__(self):
		t = '==== TS_POINTER_EVENT ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t




import enum
from typing import List
from aardwolf.commons.queuedata import RDPDATATYPE
from aardwolf.keyboard import VK_MODIFIERS

class RDP_KEYBOARD_SCANCODE:
	def __init__(self):
		self.type = RDPDATATYPE.KEYSCAN
		self.keyCode:int = None
		self.is_pressed:bool = True
		self.is_extended:bool = False
		self.modifiers:VK_MODIFIERS = VK_MODIFIERS(0)
		self.vk_code:str = None

class RDP_KEYBOARD_UNICODE:
	def __init__(self):
		self.type = RDPDATATYPE.KEYUNICODE
		self.char:str = None
		self.is_pressed:bool = None


import io
import enum

class INPUT_FLAG(enum.IntFlag):
	SCANCODES = 0x0001 #Indicates support for using scancodes in the Keyboard Event notifications (sections 2.2.8.1.1.3.1.1.1 and 2.2.8.1.2.2.1).
	MOUSEX = 0x0004 #Indicates support for Extended Mouse Event notifications (sections 2.2.8.1.1.3.1.1.4 and 2.2.8.1.2.2.4).
	FASTPATH_INPUT = 0x0008 #Advertised by RDP 5.0 and 5.1 servers to indicate support for fast-path input.
	UNICODE = 0x0010 #Indicates support for Unicode Keyboard Event notifications (sections 2.2.8.1.1.3.1.1.2 and 2.2.8.1.2.2.2).
	FASTPATH_INPUT2 = 0x0020 #Advertised by all RDP servers, except for RDP 4.0, 5.0, and 5.1 servers, to indicate support for fast-path input. Clients that do not support this flag will not be able to use fast-path input when connecting to RDP 5.2, 6.0, 6.1, 7.0, 7.1, 8.0, 8.1, 10.0, 10.1, 10.2, 10.3, 10.4, and 10.5 servers.
	UNUSED1 = 0x0040 #An unused flag that MUST be ignored by the client if it is present in the server-to-client Input Capability Set.
	UNUSED2 = 0x0080 #An unused flag that MUST be ignored by the server if it is present in the client-to-server Input Capability Set.
	MOUSE_HWHEEL = 0x0100 #Indicates support for horizontal Mouse Wheel Event notifications (sections 2.2.8.1.1.3.1.1.3 and 2.2.8.1.2.2.3).
	QOE_TIMESTAMPS = 0x0200 #Indicates support for Quality of Experience (QoE) Timestamp Event notifications (section 2.2.8.1.2.2.6). There is no slow-path support for Quality of Experience (QoE) timestamps.

class TS_INPUT_CAPABILITYSET:
	def __init__(self):
		self.inputFlags:INPUT_FLAG = INPUT_FLAG.SCANCODES
		self.pad2octets:bytes = b'\x00'*2
		self.keyboardLayout:int = 0x00000409 #US
		self.keyboardType: int = 0x00000004
		self.keyboardSubType: int = 0
		self.keyboardFunctionKey: int = 12
		self.imeFileName: str = ''

	def to_bytes(self):
		imeFileName = self.imeFileName.encode('utf-16-le').ljust(64, b'\x00')

		t = self.inputFlags.to_bytes(2, byteorder='little', signed = False)
		t += self.pad2octets
		t += self.keyboardLayout.to_bytes(4, byteorder='little', signed = False)
		t += self.keyboardType.to_bytes(4, byteorder='little', signed = False)
		t += self.keyboardSubType.to_bytes(4, byteorder='little', signed = False)
		t += self.keyboardFunctionKey.to_bytes(4, byteorder='little', signed = False)
		t += imeFileName
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return TS_INPUT_CAPABILITYSET.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = TS_INPUT_CAPABILITYSET()
		msg.inputFlags = INPUT_FLAG(int.from_bytes(buff.read(2), byteorder='little', signed = False))
		msg.pad2octets = buff.read(2)
		msg.keyboardLayout = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		msg.keyboardType = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		msg.keyboardSubType = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		msg.keyboardFunctionKey = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		try:
			msg.imeFileName = buff.read(64).decode('utf-16-le').replace('\x00', '')
		except:
			#TODO: FIX THIS (on windows XP it is some werid garbage here)
			msg.imeFileName = ''
		
		return msg

	def __repr__(self):
		t = '==== TS_INPUT_CAPABILITYSET ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t


import io
import enum

class CLIPRDR_FILECONTENTS_RESPONSE:
	def __init__(self,is_size):
		self.is_size = is_size
		self.streamId:int = None
		self.requestedFileContentsData = None #this is either a 64bit integer (if filesize was requested) OR a blob of bytes with file contents


	def to_bytes(self):
		t = self.streamId.to_bytes(4, byteorder='little', signed = False)
		t += self.requestedFileContentsData
		return t

	@staticmethod
	def from_bytes(bbuff: bytes, is_size):
		return CLIPRDR_FILECONTENTS_RESPONSE.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO, is_size):
		msg = CLIPRDR_FILECONTENTS_RESPONSE(is_size)
		msg.streamId = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		if is_size is True:
			msg.requestedFileContentsData = int.from_bytes(buff.read(8), byteorder='little', signed = False)
		else:
			msg.requestedFileContentsData = buff.read()
		return msg

	def __repr__(self):
		t = '==== CLIPRDR_FILECONTENTS_RESPONSE ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t



import uuid
import ipaddress

class FileTargetGen:
	def __init__(self, filename):
		self.filename = filename

	async def generate(self):
		try:
			with open(self.filename, 'r') as f:
				for line in f:
					line = line.strip()
					if line == '':
						continue
					yield str(uuid.uuid4()), line, None
		except Exception as e:
			yield None, None, e


class ListTargetGen:
	def __init__(self, targets):
		self.targets = targets

	async def generate(self):
		try:
			for target in self.targets:
				target = target.strip()
				try:
					ip = ipaddress.ip_address(target)
					yield str(uuid.uuid4()),str(ip), None
				except:
					try:
						for ip in ipaddress.ip_network(target, strict = False):
							yield  str(uuid.uuid4()),str(ip), None
					except:
						yield str(uuid.uuid4()), target, None
		except Exception as e:
			yield None, None, e


import io
import enum
from os import stat

from aardwolf import logger
from aardwolf.protocol.pdu.capabilities.bitmamachachehostsupp import TS_BITMAPCACHE_HOSTSUPPORT_CAPABILITYSET
from aardwolf.protocol.pdu.capabilities.bitmap import TS_BITMAP_CAPABILITYSET
from aardwolf.protocol.pdu.capabilities.bitmapcache import TS_BITMAPCACHE_CAPABILITYSET
from aardwolf.protocol.pdu.capabilities.bitmapcache2 import TS_BITMAPCACHE_CAPABILITYSET_REV2
from aardwolf.protocol.pdu.capabilities.brush import TS_BRUSH_CAPABILITYSET
from aardwolf.protocol.pdu.capabilities.bitmapcodecs import TS_BITMAPCODECS_CAPABILITYSET
from aardwolf.protocol.pdu.capabilities.colortablecache import TS_COLORTABLE_CAPABILITYSET
from aardwolf.protocol.pdu.capabilities.compdesk import TS_COMPDESK_CAPABILITYSET
from aardwolf.protocol.pdu.capabilities.control import TS_CONTROL_CAPABILITYSET
from aardwolf.protocol.pdu.capabilities.font import TS_FONT_CAPABILITYSET
from aardwolf.protocol.pdu.capabilities.frameack import TS_FRAME_ACKNOWLEDGE_CAPABILITYSET
#from aardwolf.protocol.pdu.capabilities.gdiplus import *
from aardwolf.protocol.pdu.capabilities.general import TS_GENERAL_CAPABILITYSET
from aardwolf.protocol.pdu.capabilities.glyph import TS_GLYPHCACHE_CAPABILITYSET
from aardwolf.protocol.pdu.capabilities.input import TS_INPUT_CAPABILITYSET
from aardwolf.protocol.pdu.capabilities.largepointer import TS_LARGE_POINTER_CAPABILITYSET
from aardwolf.protocol.pdu.capabilities.multifragmentupdate import TS_MULTIFRAGMENTUPDATE_CAPABILITYSET
from aardwolf.protocol.pdu.capabilities.ninegrid import TS_DRAW_NINEGRID_CAPABILITYSET
from aardwolf.protocol.pdu.capabilities.offscreen import TS_OFFSCREEN_CAPABILITYSET
from aardwolf.protocol.pdu.capabilities.order import TS_ORDER_CAPABILITYSET
from aardwolf.protocol.pdu.capabilities.pointer import TS_POINTER_CAPABILITYSET
from aardwolf.protocol.pdu.capabilities.remoteprograms import TS_REMOTE_PROGRAMS_CAP_SET
from aardwolf.protocol.pdu.capabilities.share import TS_SHARE_CAPABILITYSET
from aardwolf.protocol.pdu.capabilities.sound import TS_SOUND_CAPABILITYSET
from aardwolf.protocol.pdu.capabilities.surface import TS_SURFCMDS_CAPABILITYSET
from aardwolf.protocol.pdu.capabilities.virtualchannel import TS_VIRTUALCHANNEL_CAPABILITYSET
from aardwolf.protocol.pdu.capabilities.windowactivation import TS_WINDOWACTIVATION_CAPABILITYSET
from aardwolf.protocol.pdu.capabilities.windowlist import TS_WINDOW_LIST_CAP_SET


class CAPSTYPE(enum.Enum):
	GENERAL = 0x0001 #General Capability Set (section 2.2.7.1.1)
	BITMAP = 0x0002 #Bitmap Capability Set (section 2.2.7.1.2)
	ORDER = 0x0003 #Order Capability Set (section 2.2.7.1.3)
	BITMAPCACHE = 0x0004 #Revision 1 Bitmap Cache Capability Set (section 2.2.7.1.4.1)
	CONTROL = 0x0005 #Control Capability Set (section 2.2.7.2.2)
	ACTIVATION = 0x0007 #Window Activation Capability Set (section 2.2.7.2.3)
	POINTER = 0x0008 #Pointer Capability Set (section 2.2.7.1.5)
	SHARE = 0x0009 #Share Capability Set (section 2.2.7.2.4)
	COLORCACHE = 0x000A #Color Table Cache Capability Set ([MS-RDPEGDI] section 2.2.1.1)
	SOUND = 0x000C #Sound Capability Set (section 2.2.7.1.11)
	INPUT = 0x000D #Input Capability Set (section 2.2.7.1.6)
	FONT = 0x000E #Font Capability Set (section 2.2.7.2.5)
	BRUSH = 0x000F #Brush Capability Set (section 2.2.7.1.7)
	GLYPHCACHE = 0x0010 #Glyph Cache Capability Set (section 2.2.7.1.8)
	OFFSCREENCACHE = 0x0011 #Offscreen Bitmap Cache Capability Set (section 2.2.7.1.9)
	BITMAPCACHE_HOSTSUPPORT = 0x0012 #Bitmap Cache Host Support Capability Set (section 2.2.7.2.1)
	BITMAPCACHE_REV2 = 0x0013 #Revision 2 Bitmap Cache Capability Set (section 2.2.7.1.4.2)
	VIRTUALCHANNEL = 0x0014 #Virtual Channel Capability Set (section 2.2.7.1.10)
	DRAWNINEGRIDCACHE = 0x0015 #DrawNineGrid Cache Capability Set ([MS-RDPEGDI] section 2.2.1.2)
	DRAWGDIPLUS = 0x0016 #Draw GDI+ Cache Capability Set ([MS-RDPEGDI] section 2.2.1.3)
	RAIL = 0x0017 #Remote Programs Capability Set ([MS-RDPERP] section 2.2.1.1.1)
	WINDOW = 0x0018 #Window List Capability Set ([MS-RDPERP] section 2.2.1.1.2)
	COMPDESK = 0x0019 #Desktop Composition Extension Capability Set (section 2.2.7.2.8)
	MULTIFRAGMENTUPDATE = 0x001A #Multifragment Update Capability Set (section 2.2.7.2.6)
	LARGE_POINTER = 0x001B #Large Pointer Capability Set (section 2.2.7.2.7)
	SURFACE_COMMANDS = 0x001C #Surface Commands Capability Set (section 2.2.7.2.9)
	BITMAP_CODECS = 0x001D #Bitmap Codecs Capability Set (section 2.2.7.2.10)
	FRAME_ACKNOWLEDGE = 0x001E #Frame Acknowledge Capability Set ([MS-RDPRFX] section 2.2.1.3)

class TS_CAPS_SET:
	def __init__(self):
		self.capabilitySetType:CAPSTYPE = None
		self.lengthCapability:int = None
		self.capabilityData:bytes = None

		# high level
		self.capability = None
	
	@staticmethod
	def from_capability(cap):
		t = TS_CAPS_SET()
		t.capabilitySetType = obj2otype[type(cap)]
		t.capability = cap
		return t

	def to_bytes(self):
		cd = self.capability.to_bytes()
		t = self.capabilitySetType.value.to_bytes(2, byteorder='little', signed = False)
		t += (len(cd)+4).to_bytes(2, byteorder='little', signed = False)
		t += cd
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return TS_CAPS_SET.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = TS_CAPS_SET()
		msg.capabilitySetType = CAPSTYPE(int.from_bytes(buff.read(2), byteorder='little', signed = False))
		msg.lengthCapability = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.capabilityData = buff.read(msg.lengthCapability-4)
		
		ot = otype2obj[msg.capabilitySetType]
		if ot is not None:
			msg.capability = ot.from_bytes(msg.capabilityData)
			#if msg.capabilityData != msg.capability.to_bytes():
			#	raise Exception('mismatch!')
		else:
			logger.debug('Not implemented parser! %s' % msg.capabilitySetType)
		return msg

	def __repr__(self):
		t = '==== TS_CAPS_SET ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t

otype2obj = {
	CAPSTYPE.GENERAL : TS_GENERAL_CAPABILITYSET,
	CAPSTYPE.BITMAP : TS_BITMAP_CAPABILITYSET,
	CAPSTYPE.ORDER : TS_ORDER_CAPABILITYSET,
	CAPSTYPE.BITMAPCACHE : TS_BITMAPCACHE_CAPABILITYSET,
	CAPSTYPE.CONTROL : TS_CONTROL_CAPABILITYSET,
	CAPSTYPE.ACTIVATION : TS_WINDOWACTIVATION_CAPABILITYSET,
	CAPSTYPE.POINTER : TS_POINTER_CAPABILITYSET,
	CAPSTYPE.SHARE : TS_SHARE_CAPABILITYSET,
	CAPSTYPE.COLORCACHE : TS_COLORTABLE_CAPABILITYSET,
	CAPSTYPE.SOUND : TS_SOUND_CAPABILITYSET,
	CAPSTYPE.INPUT : TS_INPUT_CAPABILITYSET,
	CAPSTYPE.FONT : TS_FONT_CAPABILITYSET,
	CAPSTYPE.BRUSH : TS_BRUSH_CAPABILITYSET,
	CAPSTYPE.GLYPHCACHE : TS_GLYPHCACHE_CAPABILITYSET,
	CAPSTYPE.OFFSCREENCACHE : TS_OFFSCREEN_CAPABILITYSET,
	CAPSTYPE.BITMAPCACHE_HOSTSUPPORT : TS_BITMAPCACHE_HOSTSUPPORT_CAPABILITYSET,
	CAPSTYPE.BITMAPCACHE_REV2 : TS_BITMAPCACHE_CAPABILITYSET_REV2, 
	CAPSTYPE.VIRTUALCHANNEL : TS_VIRTUALCHANNEL_CAPABILITYSET,
	CAPSTYPE.DRAWNINEGRIDCACHE : TS_DRAW_NINEGRID_CAPABILITYSET,
	CAPSTYPE.DRAWGDIPLUS : None,
	CAPSTYPE.RAIL : TS_REMOTE_PROGRAMS_CAP_SET,
	CAPSTYPE.WINDOW : TS_WINDOW_LIST_CAP_SET,
	CAPSTYPE.COMPDESK : TS_COMPDESK_CAPABILITYSET,
	CAPSTYPE.MULTIFRAGMENTUPDATE : TS_MULTIFRAGMENTUPDATE_CAPABILITYSET,
	CAPSTYPE.LARGE_POINTER : TS_LARGE_POINTER_CAPABILITYSET,
	CAPSTYPE.SURFACE_COMMANDS : TS_SURFCMDS_CAPABILITYSET,
	CAPSTYPE.BITMAP_CODECS : TS_BITMAPCODECS_CAPABILITYSET,
	CAPSTYPE.FRAME_ACKNOWLEDGE : TS_FRAME_ACKNOWLEDGE_CAPABILITYSET,

}

obj2otype = {v: k for k, v in otype2obj.items()}


#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#
# Comments:
#


import ipaddress
import enum
import copy
from typing import List
from urllib.parse import urlparse, parse_qs

from asysocks.unicomm.common.target import UniTarget, UniProto
from asysocks.unicomm.common.proxy import UniProxyProto, UniProxyTarget

class RDPConnectionDialect(enum.Enum):
	RDP = 'RDP'
	VNC = 'VNC'

rdpurlconnection_param2var = {
	'RDP' : UniProto.CLIENT_TCP,
	'VNC' : UniProto.CLIENT_TCP,
}

class RDPTarget(UniTarget):
	"""
	"""
	def __init__(
			self, 
			ip:str = None, 
			port:int = 3389, 
			hostname:str = None, 
			timeout:int = 1, 
			dc_ip:int=None, 
			domain:str = None, 
			proxies:List[UniProxyTarget] = None, 
			protocol:UniProto = UniProto.CLIENT_TCP, 
			dialect:RDPConnectionDialect = RDPConnectionDialect.RDP,
			dns:str = None):
		UniTarget.__init__(self, ip, port, protocol, timeout, hostname = hostname, proxies = proxies, domain = domain, dc_ip = dc_ip, dns=dns)
		self.dialect = dialect
		if self.dialect == RDPConnectionDialect.VNC:
			self.port = 5900

	def to_target_string(self):
		return 'termsrv/%s@%s' % (self.hostname, self.domain)

	def get_copy(self, ip, port, hostname = None):
		t = RDPTarget(
			ip = ip, 
			port = port, 
			hostname = hostname, 
			timeout = self.timeout, 
			dc_ip= self.dc_ip, 
			domain = self.domain, 
			proxy = copy.deepcopy(self.proxy),
			protocol = self.protocol,
			dialect = self.dialect,
			dns=self.dns
		)
		return t

	@staticmethod
	def from_url(connection_url):
		url_e = urlparse(connection_url)
		schemes = url_e.scheme.upper().split('+')
		connection_tags = schemes[0].split('-')
		if len(connection_tags) > 1:
			dialect = RDPConnectionDialect(connection_tags[0])
			protocol = rdpurlconnection_param2var[connection_tags[1]]
		else:
			dialect = RDPConnectionDialect(connection_tags[0])
			protocol = UniProto.CLIENT_TCP
		
		if url_e.port:
			port = url_e.port
		elif dialect == RDPConnectionDialect.RDP:
			port = 3389
		elif dialect == RDPConnectionDialect.VNC:
			port = 5800
		else:
			raise Exception('Port must be provided!')
		
		unitarget, extraparams = UniTarget.from_url(connection_url, protocol, port)

		target = RDPTarget(
			ip = unitarget.ip,
			port = unitarget.port,
			hostname = unitarget.hostname,
			timeout = unitarget.timeout,
			dc_ip = unitarget.dc_ip,
			domain = unitarget.domain,
			proxies = unitarget.proxies,
			protocol = unitarget.protocol,
			dns = unitarget.dns,
			dialect = dialect,
		)
		
		return target
		
	def get_ip(self):
		if not self.ip and not self.hostname:
			raise Exception('RDPTarget must have ip or hostname defined!')
		return self.ip if self.ip is not None else self.hostname
		
	def get_hostname(self):
		return self.hostname
	
	def get_hostname_or_ip(self):
		if self.hostname:
			return self.hostname
		return self.ip
	
	def get_port(self):
		return self.port
		
	def __str__(self):
		t = '==== RDPTarget ====\r\n'
		for k in self.__dict__:
			t += '%s: %s\r\n' % (k, self.__dict__[k])
			
		return t
		
		
def test():
	s = 'TEST/victim/ntlm/nt:AAAAAAAA@10.10.10.2:3389'
	creds = RDPTarget.from_connection_string(s)
	print(str(creds))
	
	s = 'TEST/victim/sspi@10.10.10.2:3389/aaaa'
	creds = RDPTarget.from_connection_string(s)
	
	print(str(creds))
	
if __name__ == '__main__':
	test()


import io
import enum


class TS_COLORTABLE_CAPABILITYSET:
	def __init__(self):
		self.colorTableCacheSize:int = 6
		self.pad2octets:bytes = b'\x00'*2

	def to_bytes(self):
		t = self.colorTableCacheSize.to_bytes(2, byteorder='little', signed = False)
		t += self.pad2octets
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return TS_COLORTABLE_CAPABILITYSET.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = TS_COLORTABLE_CAPABILITYSET()
		msg.colorTableCacheSize = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.pad2octets = buff.read(2)
		return msg

	def __repr__(self):
		t = '==== TS_COLORTABLE_CAPABILITYSET ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t


import io
import enum

class PTRXFLAGS(enum.IntFlag):
	DOWN = 0x8000 #Indicates that a click event has occurred at the position specified by the xPos and yPos fields. The button flags indicate which button has been clicked and at least one of these flags MUST be set.
	BUTTON1 = 0x0001 #Extended mouse button 1 (also referred to as button 4) was clicked or released. If the PTRXFLAGS_DOWN flag is set, the button was clicked; otherwise, it was released.
	BUTTON2 = 0x0002 #Extended mouse button 2 (also referred to as button 5) was clicked or released. If the PTRXFLAGS_DOWN flag is set, the button was clicked; otherwise, it was released.

class TS_POINTERX_EVENT:
	def __init__(self):
		self.pointerFlags:PTRXFLAGS = None #only release flag is allowed here
		self.xPos:int = None #maybe string? #The Unicode character input code.
		self.yPos:int = None

	def to_bytes(self):
		t = self.pointerFlags.to_bytes(2, byteorder='little', signed=False)
		t += self.xPos.to_bytes(2, byteorder='little', signed=False)
		t += self.yPos.to_bytes(2, byteorder='little', signed=False)
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return TS_POINTERX_EVENT.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = TS_POINTERX_EVENT()
		msg.pointerFlags = PTRXFLAGS(int.from_bytes(buff.read(2), byteorder='little', signed=False))
		msg.xPos = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		msg.yPos = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		return msg

	def __repr__(self):
		t = '==== TS_POINTERX_EVENT ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t


import enum
import io
from aardwolf.protocol.T125.extendedinfopacket import TS_EXTENDED_INFO_PACKET

class INFO_FLAG(enum.IntFlag):
	MOUSE = 0x00000001 #Indicates that the client machine has a mouse attached.
	DISABLECTRLALTDEL = 0x00000002 #Indicates that the CTRL+ALT+DEL (or the equivalent) secure access keyboard sequence is not required at the logon prompt
	AUTOLOGON = 0x00000008 #The client requests auto logon using the included user name, password and domain.
	UNICODE = 0x00000010 #Indicates that the character set for strings in the Info Packet and Extended Info Packet (section 2.2.1.11.1.1.1) is Unicode. If this flag is absent, then the ANSI character set that is specified by the ANSI code page descriptor in the CodePage field is used for strings in the Info Packet and Extended Info Packet.
	MAXIMIZESHELL = 0x00000020 #Indicates that the alternate shell (specified in the AlternateShell field of the Info Packet structure) MUST be started in a maximized state.
	LOGONNOTIFY = 0x00000040 #Indicates that the client wants to be informed of the user name and domain used to log on to the server, as well as the ID of the session to which the user connected. The Save Session Info PDU (section 2.2.10.1) is sent from the server to notify the client of this information using a Logon Info Version 1 (section 2.2.10.1.1.1) or Logon Info Version 2 (section 2.2.10.1.1.2) structure.
	COMPRESSION = 0x00000080 #Indicates that the CompressionTypeMask is valid and contains the highest compression package type supported by the client.
	CompressionTypeMask = 0x00001E00 #Indicates the highest compression package type supported. See the discussion which follows this table for more information.
	ENABLEWINDOWSKEY = 0x00000100 #Indicates that the client uses the Windows key on Windows-compatible keyboards.
	REMOTECONSOLEAUDIO = 0x00002000 #Requests that audio played in a session hosted on a remote server be played on the server.
	FORCE_ENCRYPTED_CS_PDU = 0x00004000 #Indicates that all client-to-server traffic is encrypted when encryption is in force. Setting this flag prevents the server from processing unencrypted packets in man-in-the-middle attack scenarios. This flag is not understood by RDP 4.0, 5.0, and 5.1 servers.
	RAIL = 0x00008000 #Indicates that the remote connection being established is for the purpose of launching remote programs using the protocol defined in [MS-RDPERP] sections 2 and 3. This flag is not understood by RDP 4.0, 5.0, 5.1, and 5.2 servers.
	LOGONERRORS = 0x00010000 #Indicates a request for logon error notifications using the Save Session Info PDU. This flag is not understood by RDP 4.0, 5.0, 5.1, and 5.2 servers.
	MOUSE_HAS_WHEEL = 0x00020000 #Indicates that the mouse which is connected to the client machine has a scroll wheel. This flag is not understood by RDP 4.0, 5.0, 5.1, and 5.2 servers.
	PASSWORD_IS_SC_PIN = 0x00040000 #Indicates that the Password field in the Info Packet contains a smart card personal identification number (PIN). This flag is not understood by RDP 4.0, 5.0, 5.1, and 5.2 servers.
	NOAUDIOPLAYBACK = 0x00080000 #Indicates that audio redirection (using the protocol defined in [MS-RDPEA] sections 2 and 3) MUST NOT take place. This flag is not understood by RDP 4.0, 5.0, 5.1, and 5.2 servers. If the INFO_NOAUDIOPLAYBACK flag is not set, then audio redirection SHOULD take place if the INFO_REMOTECONSOLEAUDIO (0x00002000) flag is also not set.
	USING_SAVED_CREDS = 0x00100000 #Any user credentials sent on the wire during the RDP Connection Sequence (sections 1.3.1.1 and 1.3.1.2) have been retrieved from a credential store and were not obtained directly from the user. This flag is not understood by RDP 4.0, 5.0, 5.1, 5.2, and 6.0 servers.
	AUDIOCAPTURE = 0x00200000 #Indicates that the redirection of client-side audio input to a session hosted on a remote server is supported using the protocol defined in [MS-RDPEAI] sections 2 and 3. This flag is not understood by RDP 4.0, 5.0, 5.1, 5.2, 6.0, and 6.1 servers.
	VIDEO_DISABLE = 0x00400000 #Indicates that video redirection or playback (using the protocol defined in [MS-RDPEV] sections 2 and 3) MUST NOT take place. This flag is not understood by RDP 4.0, 5.0, 5.1, 5.2, 6.0, and 6.1 servers.
	RESERVED1 = 0x00800000 #An unused flag that is reserved for future use. This flag MUST NOT be set.
	RESERVED2 = 0x01000000 #An unused flag that is reserved for future use. This flag MUST NOT be set.
	HIDEF_RAIL_SUPPORTED = 0x02000000 #Indicates that the client supports Enhanced RemoteApp ([MS-RDPERP] section 1.3.3). The INFO_HIDEF_RAIL_SUPPORTED flag MUST be ignored if the INFO_RAIL (0x00008000) flag is not specified. Furthermore, a client that specifies the INFO_HIDEF_RAIL_SUPPORTED flag MUST send the Remote Programs Capability Set ([MS-RDPERP] section 2.2.1.1.1) to the server. The INFO_HIDEF_RAIL_SUPPORTED flag is not understood by RDP 4.0, 5.0, 5.1, 5.2, 6.0, 6.1, 7.0, 7.1, and 8.0 servers

class CompressionTypeFlag:
	TYPE_8K = 0x0 #RDP 4.0 bulk compression (section 3.1.8.4.1).
	TYPE_64K = 0x1 #RDP 5.0 bulk compression (section 3.1.8.4.2).
	TYPE_RDP6 = 0x2 #RDP 6.0 bulk compression ([MS-RDPEGDI] section 3.1.8.1).
	TYPE_RDP61 = 0x3

# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-rdpbcgr/732394f5-e2b5-4ac5-8a0a-35345386b0d1
class TS_INFO_PACKET:
	def __init__(self):
		self.CodePage:int = None
		self.flags:INFO_FLAG = None
		self.cbDomain:int = None
		self.cbUserName:int = None
		self.cbPassword:int = None
		self.cbAlternateShell:int = None
		self.cbWorkingDir:int = None
		self.Domain:str = None
		self.UserName:str = None
		self.Password:str = None
		self.AlternateShell:str = None
		self.WorkingDir:str = None
		self.extrainfo:TS_EXTENDED_INFO_PACKET = None

	def to_bytes(self):
		encoding= 'ascii'
		if INFO_FLAG.UNICODE in self.flags:
			encoding = 'utf-16-le'
		domain = (self.Domain+'\x00').encode(encoding)
		username = (self.UserName+'\x00').encode(encoding)
		password = (self.Password+'\x00').encode(encoding)
		altshell = (self.AlternateShell+'\x00').encode(encoding)
		wd = (self.WorkingDir+'\x00').encode(encoding)
		
			
		t = self.CodePage.to_bytes(4, byteorder='little', signed = False)
		t += self.flags.to_bytes(4, byteorder='little', signed = False)
		
		t += (0 if len(self.Domain) == 0 else len(domain)-2).to_bytes(2, byteorder='little', signed = False)
		t += (0 if len(self.UserName) == 0 else len(username)-2).to_bytes(2, byteorder='little', signed = False)
		t += (0 if len(self.Password) == 0 else len(password)-2).to_bytes(2, byteorder='little', signed = False)
		t += (0 if len(self.AlternateShell) == 0 else len(altshell)-2).to_bytes(2, byteorder='little', signed = False)
		t += (0 if len(self.WorkingDir) == 0 else len(wd)-2).to_bytes(2, byteorder='little', signed = False)
		t += domain
		t += username
		t += password
		t += altshell
		t += wd
		
		if self.extrainfo is not None:
			t += self.extrainfo.to_bytes()

		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return TS_INFO_PACKET.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = TS_INFO_PACKET()
		msg.CodePage = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		msg.flags = INFO_FLAG(int.from_bytes(buff.read(4), byteorder='little', signed = False))
		# NOTE: the cb.. fields contain the character count NOT the bytes count!!!! excluding the null terminator
		# also the character count depend on the UNICODE FLAG
		msg.cbDomain = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.cbUserName = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.cbPassword = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.cbAlternateShell = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.cbWorkingDir = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		encoding= 'ascii'
		if INFO_FLAG.UNICODE in msg.flags:
			encoding = 'utf-16-le'
		msg.Domain = buff.read(msg.cbDomain + 2).decode(encoding).replace('\x00', '')
		msg.UserName = buff.read(msg.cbUserName + 2).decode(encoding).replace('\x00', '')
		msg.Password = buff.read(msg.cbPassword + 2).decode(encoding).replace('\x00', '')
		msg.AlternateShell = buff.read(msg.cbAlternateShell + 2).decode(encoding).replace('\x00', '')
		msg.WorkingDir = buff.read(msg.cbWorkingDir+2).decode(encoding).replace('\x00', '')
		print(msg)
		print(buff.read(20))
		buff.seek(-20,1)
		msg.extrainfo = TS_EXTENDED_INFO_PACKET.from_buffer(buff)
		return msg

	def __repr__(self):
		t = '==== TS_INFO_PACKET ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t


import enum
import io
from aardwolf.protocol.fastpath.bitmap import TS_BITMAP_FLAG, TS_BITMAP_DATA

from aardwolf import logger
from aardwolf.commons.queuedata import RDPDATATYPE
from aardwolf.commons.queuedata.constants import VIDEO_FORMAT
from aardwolf.utils.rectconvert import rectconvert
try:
	from PIL.ImageQt import ImageQt
except ImportError:
	logger.debug('No Qt installed! Converting to qt will not work')

class RDP_VIDEO:
	def __init__(self):
		self.type = RDPDATATYPE.VIDEO
		self.x:int = None
		self.y:int = None
		self.width:int = None
		self.height:int = None
		self.bitsPerPixel:int = None
		self.is_compressed = None
		self.data:bytes = None
	
	@staticmethod
	def from_bitmapdata(bitmapdata:TS_BITMAP_DATA, output_format = VIDEO_FORMAT.QT5):
		res = RDP_VIDEO()
		res.type = RDPDATATYPE.VIDEO
		res.x = bitmapdata.destLeft
		res.y = bitmapdata.destTop
		res.width = bitmapdata.destRight - bitmapdata.destLeft + 1
		res.height = bitmapdata.destBottom - bitmapdata.destTop + 1
		res.bitsPerPixel = bitmapdata.bitsPerPixel
		res.is_compressed = TS_BITMAP_FLAG.BITMAP_COMPRESSION in bitmapdata.flags
		image_pil = rectconvert(res.width, res.height, res.bitsPerPixel, res.is_compressed, bitmapdata.bitmapDataStream)
		if output_format == VIDEO_FORMAT.PIL:
			image = image_pil

		elif output_format == VIDEO_FORMAT.RAW:
			image = image_pil.tobytes()

		elif output_format == VIDEO_FORMAT.QT5:
			image = ImageQt(image_pil)
		
		elif output_format == VIDEO_FORMAT.PNG:
			img_byte_arr = io.BytesIO()
			image_pil.save(img_byte_arr, format='PNG')
			image = img_byte_arr.getvalue()
		else:
			raise ValueError('Output format of "%s" is not supported!' % output_format)
		res.data = image
		
		return res, image_pil

	def __repr__(self):
		t = '==== RDP_VIDEO ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t



import io
import enum

# https://docs.microsoft.com/en-us/windows/win32/dataxchg/standard-clipboard-formats?redirectedfrom=MSDN
class CLIPBRD_FORMAT(enum.Enum):
	UNKNOWN = 0xFFFF
	CF_BITMAP = 2 #A handle to a bitmap (HBITMAP).
	CF_DIB = 8 #	A memory object containing a BITMAPINFO structure followed by the bitmap bits.
	CF_DIBV5 = 17 #A memory object containing a BITMAPV5HEADER structure followed by the bitmap color space information and the bitmap bits.
	CF_DIF = 5 #Software Arts' Data Interchange Format.
	CF_DSPBITMAP = 0x0082 #Bitmap display format associated with a private format. The hMem parameter must be a handle to data that can be displayed in bitmap format in lieu of the privately formatted data.
	CF_DSPENHMETAFILE = 0x008E #Enhanced metafile display format associated with a private format. The hMem parameter must be a handle to data that can be displayed in enhanced metafile format in lieu of the privately formatted data.
	CF_DSPMETAFILEPICT = 0x0083 #Metafile-picture display format associated with a private format. The hMem parameter must be a handle to data that can be displayed in metafile-picture format in lieu of the privately formatted data.
	CF_DSPTEXT = 0x0081 # Text display format associated with a private format. The hMem parameter must be a handle to data that can be displayed in text format in lieu of the privately formatted data.
	CF_ENHMETAFILE = 14 #A handle to an enhanced metafile (HENHMETAFILE).
	CF_GDIOBJFIRST = 0x0300 #Start of a range of integer values for application-defined GDI object clipboard formats. The end of the range is CF_GDIOBJLAST.
							#Handles associated with clipboard formats in this range are not automatically deleted using the GlobalFree function when the clipboard is emptied. Also, when using values in this range, the hMem parameter is not a handle to a GDI object, but is a handle allocated by the GlobalAlloc function with the GMEM_MOVEABLE flag.
	CF_GDIOBJLAST = 0x03FF #See CF_GDIOBJFIRST.
	CF_HDROP = 15 #A handle to type HDROP that identifies a list of files. An application can retrieve information about the files by passing the handle to the DragQueryFile function.
	CF_LOCALE = 16  #The data is a handle (HGLOBAL) to the locale identifier (LCID) associated with text in the clipboard. When you close the clipboard, if it contains CF_TEXT data but no CF_LOCALE data, the system automatically sets the CF_LOCALE format to the current input language. You can use the CF_LOCALE format to associate a different locale with the clipboard text.
					#An application that pastes text from the clipboard can retrieve this format to determine which character set was used to generate the text.
					#Note that the clipboard does not support plain text in multiple character sets. To achieve this, use a formatted text data type such as RTF instead.
					#The system uses the code page associated with CF_LOCALE to implicitly convert from CF_TEXT to CF_UNICODETEXT. Therefore, the correct code page table is used for the conversion.
	CF_METAFILEPICT = 3 #Handle to a metafile picture format as defined by the METAFILEPICT structure. When passing a CF_METAFILEPICT handle by means of DDE, the application responsible for deleting hMem should also free the metafile referred to by the CF_METAFILEPICT handle.
	CF_OEMTEXT = 7 #Text format containing characters in the OEM character set. Each line ends with a carriage return/linefeed (CR-LF) combination. A null character signals the end of the data.
	CF_OWNERDISPLAY = 0x0080 #Owner-display format. The clipboard owner must display and update the clipboard viewer window, and receive the WM_ASKCBFORMATNAME, WM_HSCROLLCLIPBOARD, WM_PAINTCLIPBOARD, WM_SIZECLIPBOARD, and WM_VSCROLLCLIPBOARD messages. The hMem parameter must be NULL.
	CF_PALETTE = 9 #Handle to a color palette. Whenever an application places data in the clipboard that depends on or assumes a color palette, it should place the palette on the clipboard as well.
					# If the clipboard contains data in the CF_PALETTE (logical color palette) format, the application should use the SelectPalette and RealizePalette functions to realize (compare) any other data in the clipboard against that logical palette.
					# When displaying clipboard data, the clipboard always uses as its current palette any object on the clipboard that is in the CF_PALETTE format.
	CF_PENDATA = 10 #Data for the pen extensions to the Microsoft Windows for Pen Computing.
	CF_PRIVATEFIRST = 0x0200 #Start of a range of integer values for private clipboard formats. The range ends with CF_PRIVATELAST. Handles associated with private clipboard formats are not freed automatically; the clipboard owner must free such handles, typically in response to the WM_DESTROYCLIPBOARD message.
	CF_PRIVATELAST = 0x02FF #	See CF_PRIVATEFIRST.
	CF_RIFF = 11 #Represents audio data more complex than can be represented in a CF_WAVE standard wave format.
	CF_SYLK = 4 #	Microsoft Symbolic Link (SYLK) format.
	CF_TEXT = 1 #Text format. Each line ends with a carriage return/linefeed (CR-LF) combination. A null character signals the end of the data. Use this format for ANSI text.
	CF_TIFF = 6 #	Tagged-image file format.
	CF_UNICODETEXT = 13 #Unicode text format. Each line ends with a carriage return/linefeed (CR-LF) combination. A null character signals the end of the data.
	CF_WAVE = 12 #Represents audio data in one of the standard wave formats, such as 11 kHz or 22 kHz PCM.
	
	# this was found over the interwebs and kinda important to have...
	CF_FILENAME = 0xc006
	CF_FILENAMEW = 0xc007

"""
[3748] Dialog HWND is 202CC
[3748] Clipboard format is 0xc0b7, Name is Shell IDList Array
[3748] Clipboard format is 0xc0c1, Name is Preferred DropEffect
[3748] Clipboard format is 0xc27c, Name is UsingDefaultDragImage
[3748] Clipboard format is 0xc27d, Name is DragImageBits
[3748] Clipboard format is 0xc0c7, Name is DragContext
[3748] Clipboard format is 0xc0c6, Name is InShellDragLoop
[3748] Clipboard format is 0x0f, Name is Predefined Clipboard format
[3748] Clipboard format is 0xc006, Name is FileName
[3748] Clipboard format is 0xc007, Name is FileNameW
[3748] Clipboard format is 0xc134, Name is IsShowingLayered
[3748] Clipboard format is 0xc133, Name is DragWindow
[3748] Clipboard format is 0xc281, Name is IsComputingImage
[3748] Clipboard format is 0xc0d0, Name is DropDescription
[3748] Clipboard format is 0xc0d1, Name is DisableDragText
[3748] Clipboard format is 0xc280, Name is IsShowingText
[3748] Clipboard format is 0xc284, Name is ComputedDragImage
"""



class CLIPRDR_FORMAT_LIST:
	def __init__(self, longnames:bool = False, encoding:str = 'utf-16-le'):
		self.templist:list = []

	def to_bytes(self):
		t = b''
		for fm in self.templist:
			t += fm.to_bytes()
		return t

	@staticmethod
	def from_bytes(bbuff: bytes, longnames:bool = False, encoding:str = 'utf-16-le'):
		return CLIPRDR_FORMAT_LIST.from_buffer(io.BytesIO(bbuff), longnames = longnames, encoding = encoding)

	@staticmethod
	def from_buffer(buff: io.BytesIO, longnames:bool = False, encoding:str = 'utf-16-le'):
		obj = CLIPRDR_SHORT_FORMAT_NAME
		if longnames is True:
			obj = CLIPRDR_LONG_FORMAT_NAME

		msg = CLIPRDR_FORMAT_LIST(longnames, encoding)
		# this is because of the terrible protocol design...
		for _ in range(255):
			if buff.read(1) == b'':
				break
			buff.seek(-1,1)
			msg.templist.append(obj.from_buffer(buff))
		return msg

	def __repr__(self):
		t = '==== CLIPRDR_FORMAT_LIST ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t


class CLIPRDR_SHORT_FORMAT_NAME:
	def __init__(self, encoding = 'utf-16-le'):
		self.formatId:CLIPBRD_FORMAT = None
		self.formatName:str = '' #always 32 bytes total, truncated if need be
		self.encoding = encoding

	def to_bytes(self):
		t = self.formatId.value.to_bytes(4, byteorder='little', signed=False)
		if self.encoding == 'ascii':
			fn = self.formatName[:32]
		else:
			fn = self.formatName[:16]
		t += fn.encode(self.encoding).ljust(32, b'\x00')
		return t

	@staticmethod
	def from_bytes(bbuff: bytes, encoding:str = 'utf-16-le'):
		return CLIPRDR_SHORT_FORMAT_NAME.from_buffer(io.BytesIO(bbuff), encoding = encoding)

	@staticmethod
	def from_buffer(buff: io.BytesIO, encoding:str='utf-16-le'):
		msg = CLIPRDR_SHORT_FORMAT_NAME(encoding)
		msg.formatId = CLIPBRD_FORMAT(int.from_bytes(buff.read(4), byteorder='little', signed=False))
		msg.formatName = buff.read(32).decode(encoding).replace('\x00', '')
		return msg

	def __repr__(self):
		t = '==== CLIPRDR_SHORT_FORMAT_NAME ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t


class CLIPRDR_LONG_FORMAT_NAME:
	def __init__(self, encoding = 'utf-16-le'):
		self.formatId:CLIPBRD_FORMAT = None
		self.wszFormatName:str = None #variable, if none or '' then one single b'\x00'

	def to_bytes(self):
		if self.wszFormatName is None or self.wszFormatName == '':
			fn = b'\x00'
		else:
			fn = self.wszFormatName.encode('utf-16-le')
		if isinstance(self.formatId, int):
			t = self.formatId.to_bytes(4, byteorder='little', signed=False)
		else:
			t = self.formatId.value.to_bytes(4, byteorder='little', signed=False)
		t += fn
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return CLIPRDR_LONG_FORMAT_NAME.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = CLIPRDR_LONG_FORMAT_NAME()
		t = buff.read(4)
		# not all formats are documented!!!!
		try:
			msg.formatId = CLIPBRD_FORMAT(int.from_bytes(t, byteorder='little', signed=False))
		except Exception as e:
			msg.formatId = int.from_bytes(t, byteorder='little', signed=False)
		
		t = buff.read(1)
		if t == b'\x00':
			msg.wszFormatName = ''
		else:
			# this is arbitrary, thank you documentation...
			for _ in range(255):
				t += buff.read(1)
				if t.find(b'\x00\x00\x00'):
					break	
			msg.wszFormatName = t.decode('utf-16-le').replace('\x00', '')
		return msg

	def __repr__(self):
		t = '==== CLIPRDR_LONG_FORMAT_NAME ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t


from aardwolf.protocol.fastpath import TS_FP_UPDATE_PDU
import rle                                                                 
from PyQt5.QtGui import QImage, QTransform

def RDPBitmapToQtImage(width, height, bitsPerPixel, isCompress, data):
	"""
	@summary: Bitmap transformation to Qt object
	@param width: width of bitmap
	@param height: height of bitmap
	@param bitsPerPixel: number of bit per pixel
	@param isCompress: use RLE compression
	@param data: bitmap data
	"""
	image = None
	#allocate
	
	if bitsPerPixel == 15:
		if isCompress:
			buf = bytes(width * height * 2)
			rle.bitmap_decompress(buf, width, height, data, 2)
			image = QImage(buf, width, height, QImage.Format_RGB555)
		else:
			image = QImage(data, width, height, QImage.Format_RGB555).transformed(QTransform(1.0, 0.0, 0.0, -1.0, 0.0, 0.0))
	
	elif bitsPerPixel == 16:
		if isCompress:
			buf = bytes(width * height * 2)
			rle.bitmap_decompress(buf, width, height, data, 2)
			image = QImage(buf, width, height, QImage.Format_RGB16)
		else:
			image = QImage(data, width, height, QImage.Format_RGB16).transformed(QTransform(1.0, 0.0, 0.0, -1.0, 0.0, 0.0))
	
	elif bitsPerPixel == 24:
		if isCompress:
			buf = bytes(width * height * 3)
			rle.bitmap_decompress(buf, width, height, data, 3)
			image = QImage(buf, width, height, QImage.Format_RGB888)
		else:
			image = QImage(data, width, height, QImage.Format_RGB888).transformed(QTransform(1.0, 0.0, 0.0, -1.0, 0.0, 0.0))
			
	elif bitsPerPixel == 32:
		if isCompress:
			buf = bytes(width * height * 4)
			rle.bitmap_decompress(buf, width, height, data, 4)
			image = QImage(buf, width, height, QImage.Format_RGB32)
		else:
			image = QImage(data, width, height, QImage.Format_RGB32).transformed(QTransform(1.0, 0.0, 0.0, -1.0, 0.0, 0.0))
	else:
		print("Receive image in bad format")
		image = QImage(width, height, QImage.Format_RGB32)
	return image



from aardwolf.commons.connection.url import RDPConnectionURL
from aardwolf.commons.connection.credential import RDPCredentialTypes

RDP_CONNECTION_PARAM_MANDATORY = ['domain','user','password','auth_type', 'timeout', 'protocol']


class RDPConnectionParams:
	def __init__(self):
		pass
	
	@staticmethod
	def extend_parser(parser):
		group = parser.add_argument_group('Parameter based connection')
		group.add_argument('-d', '--domain', default='.', help='Domainname')
		group.add_argument('-u', '--user', help='Username')
		group.add_argument('-p', '--password', help='Password (or other secret for authentication)')
		group.add_argument('-a', '--auth-type', choices=[e.value for e in RDPCredentialTypes], default='ntlm-password', help='authentication type')
		group.add_argument('-t', '--timeout', type=int, default=5, help='Connection timeout')
		group.add_argument('--protocol', choices=['RDP', 'RDP2', 'RDP3'], default='RDP2', help='RDP protocol version')
		group.add_argument('--dc', help='DC ip for kerberos')

	@staticmethod
	def parse_args(args):
		for p in RDP_CONNECTION_PARAM_MANDATORY:
			if hasattr(args, p) is False:
				raise Exception('The parameter "%s" is mandatory for parameter based connection setup!' % p)
			if getattr(args, p) is None:
				raise Exception('The parameter "%s" cannot be empty!' % p)

		url_template = '{protocol}+{auth_type}://{domainname}\\{username}:{password}@{host}/'
		params = {
			'protocol'  : args.protocol,
			'auth_type' : args.auth_type,
			'domainname': args.domain,
			'username'  : args.user,
			'password'  : args.password,
			'host'      : args.host if hasattr(args, 'host') else None,
		}

		url = url_template.format(**params)
		print(url)
		return url


if __name__ == '__main__':
	import argparse
	parser = argparse.ArgumentParser(prog='TEST')

	RDPConnectionParams.extend_parser(parser)

	print(parser)

	args = parser.parse_args()

	RDPConnectionParams.parse_args(args)



import typing


from asysocks.unicomm.common.connection import UniConnection

from aardwolf.protocol.x224.constants import TPDU_TYPE, NEG_FLAGS, SUPP_PROTOCOLS
from aardwolf.protocol.x224.client.connectionrequest import ConnectionRequest
from aardwolf.protocol.x224.server.connectionconfirm import ConnectionConfirm
from aardwolf.protocol.x224.data import Data
from aardwolf.protocol.x224 import X224Packet
from aardwolf.protocol.x224.constants import NEG_FLAGS, SUPP_PROTOCOLS
from aardwolf.protocol.x224.client.connectionrequest import ConnectionRequest, RDP_NEG_REQ
from aardwolf.protocol.x224.server.connectionconfirm import ConnectionConfirm


class X224Network:
	def __init__(self, connection:UniConnection):
		self.connection = connection
	
	async def client_negotiate(self, flags:NEG_FLAGS = 0, supported_protocols:SUPP_PROTOCOLS = SUPP_PROTOCOLS.SSL|SUPP_PROTOCOLS.HYBRID_EX, to_raise = True):
		reply = None
		try:
			negreq = RDP_NEG_REQ()
			negreq.flags = flags
			negreq.requestedProtocols = supported_protocols
			
			conreq = ConnectionRequest()
			conreq.SRC_REF = 0
			conreq.cookie = b'Cookie: mstshash=devel\r\n'
			conreq.rdpNegReq = negreq

			await self.connection.write(conreq.to_bytes())
			reply = await self.read()
			if reply.CR != TPDU_TYPE.CONNECTION_CONFIRM:
				raise Exception('Server sent back unknown TPDU type! %s' % reply.CR)
			reply = typing.cast(ConnectionConfirm, reply)
			if reply.rdpNegData is None:
				return reply, None
			
			if reply.rdpNegData.type == 3 and to_raise is True:
				#server denied our request!
				raise Exception('Server refused our connection request! Reason: %s' % reply.rdpNegData.failureCode.name)

			return reply, None
		except Exception as e:
			return reply, e

	async def read(self):
		is_fastpath, tpktdata = await self.connection.read_one()
		if is_fastpath is True:
			raise Exception('Fastpath packet should never be here!')
		return X224Packet.from_bytes(tpktdata)

	async def write(self, data):
		msg = Data()
		msg.data = data
		await self.connection.write(msg.to_bytes())

			
			


import copy
from asysocks.unicomm.common.scanner.common import *
from aardwolf.commons.factory import RDPConnectionFactory
from aardwolf.protocol.x224.constants import SUPP_PROTOCOLS


class RDPLoginRes:
	def __init__(self, success):
		self.success = success

	def get_header(self):
		return ['success']

	def to_line(self, separator = '\t'):
		return str(self.success)

class RDPLoginScanner:
	def __init__(self, factory:RDPConnectionFactory):
		self.factory:RDPConnectionFactory = factory
		self.flags_test = [SUPP_PROTOCOLS.HYBRID_EX, None]

	async def run(self, targetid, target, out_queue):
		try:
			for i, proto in enumerate(self.flags_test):
				result = 'TRUE'
				if proto == None:
					result = 'MAYBE'
				ios = self.factory.get_settings()
				ios.supported_protocols = proto
				async with self.factory.create_connection_newtarget(target, ios) as connection:
					_, err = await connection.connect()
					if err is not None:
						result = 'FALSE'
					
					if result == 'FALSE' and i != (len(self.flags_test)-1):
						# avoid printing NO multiple times
						continue
					
					await out_queue.put(ScannerData(target, RDPLoginRes(result)))
					if result == 'TRUE':
						break

		except Exception as e:
			await out_queue.put(ScannerError(target, e))




import asyncio
import logging
from aardwolf.commons.factory import RDPConnectionFactory
from aardwolf.commons.queuedata.constants import VIDEO_FORMAT
from aardwolf.commons.iosettings import RDPIOSettings
from asysocks.unicomm.common.scanner.targetgen import UniTargetGen
from asysocks.unicomm.common.scanner.scanner import UniScanner
from aardwolf import logger

from aardwolf.examples.scanners.rdplogin import RDPLoginScanner
from aardwolf.examples.scanners.rdpscaps import RDPCapabilitiesScanner
from aardwolf.examples.scanners.rdpscreen import RDPScreenshotScanner

rdpscan_options = {
	'login' : (RDPLoginScanner,'Checks if user can log in to hosts'),
	'caps'  : (RDPCapabilitiesScanner,'Lists RDP connection flags available'),
	'screen': (RDPScreenshotScanner,'Takes screenshot of the remote session'),
}

async def amain():
	import argparse

	scannertpes_usage = '\r\nall: Runs all scanners\r\n'
	for k in rdpscan_options:
		scannertpes_usage += '    %s: %s\r\n' % (k, rdpscan_options[k][1])
	
	usage = """
Scanner types (-s param):
    %s
"""% scannertpes_usage

	parser = argparse.ArgumentParser(description='RDP scanner', usage=usage)
	parser.add_argument('-w', '--worker-count', type=int, default=100, help='Parallell count')
	parser.add_argument('-t', '--timeout', type=int, default=10, help='Timeout for each connection')
	parser.add_argument('--no-progress', action='store_false', help='Disable progress bar')
	parser.add_argument('-o', '--out-file', help='Output file path.')
	parser.add_argument('-s', '--scan', action='append', required=True, help='Scanner type')
	parser.add_argument('-e', '--errors', action='store_true', help='Includes errors in output. It will mess up the formatting!')
	parser.add_argument('url', help = 'Connection string in URL format')
	parser.add_argument('targets', nargs='*', help = 'Hostname or IP address or file with a list of targets')
	args = parser.parse_args()

	if len(args.targets) == 0:
		print('No targets defined!')
		return
	
	logger.setLevel(logging.INFO)

	iosettings = RDPIOSettings()
	iosettings.channels = []
	iosettings.video_out_format = VIDEO_FORMAT.RAW
	iosettings.clipboard_use_pyperclip = False
	
	connectionfactory = RDPConnectionFactory.from_url(args.url, iosettings)
	scantypes = []
	for x in args.scan:
		scantypes.append(x.lower())
	executors = []
	if 'all' in scantypes:
		for k in rdpscan_options:
			executors.append(rdpscan_options[k][0](connectionfactory))
	else:
		for scantype in scantypes:
			if scantype not in rdpscan_options:
				print('Unknown scan type: "%s"' % scantype)
				return
			executors.append(rdpscan_options[scantype][0](connectionfactory))
		
	tgen = UniTargetGen.from_list(args.targets)
	scanner = UniScanner('RDPScanner', executors, [tgen], worker_count=args.worker_count, host_timeout=args.timeout)
	await scanner.scan_and_process(progress=args.no_progress, out_file=args.out_file, include_errors=args.errors)

def main():
	asyncio.run(amain())

if __name__ == '__main__':
	main()


import asyncio
import io
import base64
import datetime
from asysocks.unicomm.common.scanner.common import *
from aardwolf.commons.factory import RDPConnectionFactory
from aardwolf.protocol.x224.constants import SUPP_PROTOCOLS
from aardwolf.commons.queuedata.constants import MOUSEBUTTON, VIDEO_FORMAT
from aardwolf.commons.iosettings import RDPIOSettings


class RDPScreenshotRes:
	def __init__(self, target, screendata):
		self.target = target
		self.screendata = screendata

	def get_fname(self):
		return 'rdpscreen_%s_%s.png' % (self.target, datetime.datetime.now().strftime("%Y%m%d_%H%M%S"))

	def get_fdata(self):
		return self.screendata

	def get_header(self):
		return ['screendata']

	def to_line(self, separator = '\t'):
		return base64.b64encode(self.screendata).decode()

class RDPScreenshotScanner:
	def __init__(self, factory:RDPConnectionFactory):
		self.factory:RDPConnectionFactory = factory
		self.screentime = 5
		self.iosettings = RDPIOSettings()
		self.iosettings.channels = []
		self.iosettings.video_width = 1024
		self.iosettings.video_height = 768
		self.iosettings.video_bpp_min = 15 #servers dont support 8 any more :/
		self.iosettings.video_bpp_max = 32
		self.iosettings.video_out_format = VIDEO_FORMAT.PNG #PIL produces incorrect picture for some reason?! TODO: check bug
		self.iosettings.clipboard_use_pyperclip = False


	async def run(self, targetid, target, out_queue):
		connection = None
		try:
			connection = self.factory.create_connection_newtarget(target, self.iosettings)
			_, err = await connection.connect()
			if err is not None:
				raise err
			
			await asyncio.sleep(self.screentime)

		except asyncio.CancelledError:
			return
		except Exception as e:
			raise
		finally:
			if connection is not None and connection.desktop_buffer_has_data is True:
				buffer = connection.get_desktop_buffer(VIDEO_FORMAT.PIL)
				data = io.BytesIO()
				buffer.save(data,'png')
				data.seek(0,0)
				await out_queue.put(ScannerData(target, RDPScreenshotRes(target, data.read())))
			if connection is not None:
				await connection.terminate()


import io
import enum

from aardwolf.extensions.RDPECLIP.protocol.clipboardcapabilities import CLIPRDR_CAPS
from aardwolf.extensions.RDPECLIP.protocol.clienttemporarydirectory import CLIPRDR_TEMP_DIRECTORY
from aardwolf.extensions.RDPECLIP.protocol.lockclipboarddata import CLIPRDR_LOCK_CLIPDATA
from aardwolf.extensions.RDPECLIP.protocol.unlockclipboarddata import CLIPRDR_UNLOCK_CLIPDATA
from aardwolf.extensions.RDPECLIP.protocol.formatdatarequest import CLIPRDR_FORMAT_DATA_REQUEST
from aardwolf.extensions.RDPECLIP.protocol.formatlist import CLIPRDR_FORMAT_LIST
from aardwolf.extensions.RDPECLIP.protocol.formatdataresponse import CLIPRDR_FORMAT_DATA_RESPONSE
from aardwolf.extensions.RDPECLIP.protocol.filecontentsrequest import CLIPRDR_FILECONTENTS_REQUEST
from aardwolf.extensions.RDPECLIP.protocol.filecontentsresponse import CLIPRDR_FILECONTENTS_RESPONSE


class CB_FLAG(enum.IntFlag):
	CB_RESPONSE_OK = 0x0001 #Used by the Format List Response PDU, Format Data Response PDU, and File Contents Response PDU to indicate that the associated request Format List PDU, Format Data Request PDU, and File Contents Request PDU were processed successfully.
	CB_RESPONSE_FAIL = 0x0002 #Used by the Format List Response PDU, Format Data Response PDU, and File Contents Response PDU to indicate that the associated Format List PDU, Format Data Request PDU, and File Contents Request PDU were not processed successfully.
	CB_ASCII_NAMES = 0x0004 #Used by the Short Format Name variant of the Format List Response PDU to indicate that the format names are in ASCII 8.

class CB_TYPE(enum.Enum):
	CB_MONITOR_READY = 0x0001 #Monitor Ready PDU
	CB_FORMAT_LIST = 0x0002 #Format List PDU
	CB_FORMAT_LIST_RESPONSE = 0x0003 #Format List Response PDU
	CB_FORMAT_DATA_REQUEST = 0x0004 #Format Data Request PDU
	CB_FORMAT_DATA_RESPONSE = 0x0005 #Format Data Response PDU
	CB_TEMP_DIRECTORY = 0x0006 #Temporary Directory PDU
	CB_CLIP_CAPS = 0x0007 #Clipboard Capabilities PDU
	CB_FILECONTENTS_REQUEST = 0x0008 #File Contents Request PDU
	CB_FILECONTENTS_RESPONSE = 0x0009 #File Contents Response PDU
	CB_LOCK_CLIPDATA = 0x000A #Lock Clipboard Data PDU
	CB_UNLOCK_CLIPDATA = 0x000B #Unlock Clipboard Data PDU

class CLIPRDR_HEADER:
	def __init__(self):
		self.msgType:CB_TYPE = None
		self.msgFlags:CB_FLAG = None
		self.dataLen:int = None

	def to_bytes(self):
		t = self.msgType.value.to_bytes(2, byteorder='little', signed = False)
		t += self.msgFlags.to_bytes(2, byteorder='little', signed = False)
		t += self.dataLen.to_bytes(4, byteorder='little', signed = False)
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return CLIPRDR_HEADER.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = CLIPRDR_HEADER()
		msg.msgType = CB_TYPE(int.from_bytes(buff.read(2), byteorder='little', signed = False))
		msg.msgFlags = CB_FLAG(int.from_bytes(buff.read(2), byteorder='little', signed = False))
		msg.dataLen = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		return msg

	@staticmethod
	def parse_packet_bytes(bbuff: bytes):
		return CLIPRDR_HEADER.parse_packet_buffer(io.BytesIO(bbuff))
	
	@staticmethod
	def parse_packet_buffer(buff: bytes):
		hdr = CLIPRDR_HEADER.from_buffer(buff)
		data = buff.read(hdr.dataLen)
		if type2obj[hdr.msgType] is None:
			return hdr, None
		return hdr, type2obj[hdr.msgType].from_bytes(data)
	
	@staticmethod
	def serialize_packet(msgType:CB_TYPE, msgFlags:CB_FLAG, obj):
		data = obj
		if obj is not None:
			if isinstance(obj, bytes):
				data = obj
			else:
				data = obj.to_bytes()
		hdr = CLIPRDR_HEADER()
		hdr.msgType = msgType
		hdr.msgFlags = msgFlags
		if obj is not None:
			hdr.dataLen = len(data)
			return hdr.to_bytes() + data
		else:
			hdr.dataLen = 0
			return hdr.to_bytes()
		
		


	def __repr__(self):
		t = '==== CLIPRDR_HEADER ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t

# not all object here are parsable automatically without outside information given to the respective parser!
type2obj = {
	CB_TYPE.CB_MONITOR_READY : None, #normal
	CB_TYPE.CB_FORMAT_LIST : CLIPRDR_FORMAT_LIST,
	CB_TYPE.CB_FORMAT_LIST_RESPONSE : None, #normal
	CB_TYPE.CB_FORMAT_DATA_REQUEST : CLIPRDR_FORMAT_DATA_REQUEST,
	CB_TYPE.CB_FORMAT_DATA_RESPONSE : CLIPRDR_FORMAT_DATA_RESPONSE, 
	CB_TYPE.CB_TEMP_DIRECTORY : CLIPRDR_TEMP_DIRECTORY,
	CB_TYPE.CB_CLIP_CAPS : CLIPRDR_CAPS,
	CB_TYPE.CB_FILECONTENTS_REQUEST : CLIPRDR_FILECONTENTS_REQUEST,
	CB_TYPE.CB_FILECONTENTS_RESPONSE : CLIPRDR_FILECONTENTS_RESPONSE,
	CB_TYPE.CB_LOCK_CLIPDATA : CLIPRDR_LOCK_CLIPDATA,
	CB_TYPE.CB_UNLOCK_CLIPDATA : CLIPRDR_UNLOCK_CLIPDATA,
}



import io
import enum

class TS_SYNC(enum.IntFlag):
	SCROLL_LOCK = 0x00000001 #Indicates that the Scroll Lock indicator light SHOULD be on.
	NUM_LOCK = 0x00000002 #Indicates that the Num Lock indicator light SHOULD be on.
	CAPS_LOCK = 0x00000004 #Indicates that the Caps Lock indicator light SHOULD be on.
	KANA_LOCK = 0x00000008 #Indicates that the Kana Lock indicator light SHOULD be on.

class TS_SYNC_EVENT:
	def __init__(self):
		self.pad2Octets:bytes = None
		self.toggleFlags:TS_SYNC = None 

	def to_bytes(self):
		t = self.pad2Octets
		t += self.toggleFlags.to_bytes(2, byteorder='little', signed=False)
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return TS_SYNC_EVENT.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = TS_SYNC_EVENT()
		msg.pad2Octets = buff.read(2)
		msg.toggleFlags = TS_SYNC(int.from_bytes(buff.read(2), byteorder='little', signed=False))
		return msg

	def __repr__(self):
		t = '==== TS_SYNC_EVENT ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t


import io
import enum


class TS_WINDOWACTIVATION_CAPABILITYSET:
	def __init__(self):
		self.helpKeyFlag:bool = False
		self.helpKeyIndexFlag:bool = False
		self.helpExtendedKeyFlag:bool = False
		self.windowManagerKeyFlag:bool = False

	def to_bytes(self):
		t = int(self.helpKeyFlag).to_bytes(2, byteorder='little', signed = False)
		t += int(self.helpKeyIndexFlag).to_bytes(2, byteorder='little', signed = False)
		t += int(self.helpExtendedKeyFlag).to_bytes(2, byteorder='little', signed = False)
		t += int(self.windowManagerKeyFlag).to_bytes(2, byteorder='little', signed = False)
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return TS_WINDOWACTIVATION_CAPABILITYSET.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = TS_WINDOWACTIVATION_CAPABILITYSET()
		msg.helpKeyFlag = bool(int.from_bytes(buff.read(2), byteorder='little', signed = False))
		msg.helpKeyIndexFlag = bool(int.from_bytes(buff.read(2), byteorder='little', signed = False))
		msg.helpExtendedKeyFlag = bool(int.from_bytes(buff.read(2), byteorder='little', signed = False))
		msg.windowManagerKeyFlag = bool(int.from_bytes(buff.read(2), byteorder='little', signed = False))	
		return msg

	def __repr__(self):
		t = '==== TS_WINDOWACTIVATION_CAPABILITYSET ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t


import asyncio

import logging
import traceback
from aardwolf.protocol.tpkt import TPKT
from aardwolf.protocol.x224 import X224Packet
from aardwolf.protocol.x224.data import Data
from asysocks.unicomm.common.packetizers import Packetizer

class CredSSPPacketizer(Packetizer):
	def __init__(self):
		Packetizer.__init__(self, 65535)
		self.in_buffer = b''
	
	def flush_buffer(self):
		data = self.in_buffer
		self.in_buffer = b''
		return data

	@staticmethod
	def calcualte_length(data):
		if data[1] <= 127:
			return data[1] + 2
		else:
			bcount = data[1] - 128
			return int.from_bytes(data[2:2+bcount], byteorder = 'big', signed = False) + bcount + 2
	
	def process_buffer(self):
		preread = 6
		remaining_length = -1
		while True:
			if len(self.in_buffer) < preread:
				break
			lb = self.in_buffer[:preread]
			remaining_length = CredSSPPacketizer.calcualte_length(lb) - preread
			if len(self.in_buffer) >= remaining_length+preread:
				data = self.in_buffer[:remaining_length+preread]
				self.in_buffer = self.in_buffer[remaining_length+preread:]
				yield data
				continue
			break
		

	async def data_out(self, data):
		yield data

	async def data_in(self, data):
		if data is None:
			yield data
		self.in_buffer += data
		for packet in self.process_buffer():
			yield packet

class TPKTPacketizer(Packetizer):
	def __init__(self):
		Packetizer.__init__(self)
		self.wrap_x224 = False
		self.in_buffer = b''
	
	def flush_buffer(self):
		data = self.in_buffer
		self.in_buffer = b''
		return data

	def packetizer_control(self, cmd):
		if cmd == 'X224':
			self.wrap_x224 = True
	
	def process_buffer(self):
		msgsize = None
		is_fastpath = False
		while True:
			if len(self.in_buffer) < 4:
				break
			if self.in_buffer[0] == 3:
				msgsize = int.from_bytes(self.in_buffer[2:4], byteorder='big', signed=False)
			else:
				is_fastpath = True
				msgsize = self.in_buffer[1]
				if msgsize >> 7 == 1:
					msgsize = int.from_bytes(self.in_buffer[1:3], byteorder='big', signed=False)
					msgsize = msgsize & 0x7fff
			
			if len(self.in_buffer)>=msgsize:
				if is_fastpath is False:
					msg = TPKT.from_bytes(self.in_buffer[:msgsize])
					self.in_buffer = self.in_buffer[msgsize:]
					msgsize = None
					if self.wrap_x224 is True:
						yield (False, X224Packet.from_bytes(msg.tpdu))
					else:
						yield (False, msg.tpdu)
					
				else:
					res = self.in_buffer[:msgsize]
					self.in_buffer = self.in_buffer[msgsize:]
					msgsize = None
					is_fastpath = False
					yield (True, res)

				if len(self.in_buffer) > 0:
					continue
			break

	async def data_out(self, data):
		tpkt = TPKT()
		tpkt.tpdu = data
		#print('TPKT -> %s' % tpkt.to_bytes())
		yield tpkt.to_bytes()

	async def data_in(self, data):
		if data is None:
			yield data
		self.in_buffer += data
		for packet in self.process_buffer():
			yield packet




import io
import enum

from aardwolf.protocol.T124.userdata.constants import TS_UD_TYPE, ENCRYPTION_FLAG, ENCRYPTION_LEVEL
from asn1crypto.x509 import Certificate

class CERT_CHAIN_VERSION(enum.Enum):
	VERSION_0 = 0x00000000 # same as ver1
	VERSION_1 = 0x00000001 #The certificate contained in the certData field is a Server Proprietary Certificate (section 2.2.1.4.3.1.1).
	VERSION_2 = 0x00000002

# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-rdpbcgr/54e72cc6-3422-404c-a6b4-2486db125342
class SERVER_CERTIFICATE:
	def __init__(self):
		self.certChainVersion:int = None
		self.t:bool = None
		self.certData:bytes = None

		self.certificate = None
		self.exponent:int = None
		self.modulus:int = None
		self.bitlen:int = None
		
	def to_bytes(self):
		raise NotImplementedError()

	@staticmethod
	def from_bytes(bbuff: bytes):
		return SERVER_CERTIFICATE.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = SERVER_CERTIFICATE()
		temp = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		msg.certChainVersion = CERT_CHAIN_VERSION((temp << 1) >> 1)
		msg.t = bool(temp >> 31)
		if msg.certChainVersion in [CERT_CHAIN_VERSION.VERSION_1, CERT_CHAIN_VERSION.VERSION_0]:
			msg.certificate = PROPRIETARYSERVERCERTIFICATE.from_buffer(buff)
			msg.exponent = msg.certificate.PublicKeyBlob.pubExp
			msg.modulus = int.from_bytes(msg.certificate.PublicKeyBlob.modulus, byteorder='little', signed=False)
			msg.bitlen = msg.certificate.PublicKeyBlob.bitlen
		else:
			msg.certificate = Certificate.load(buff.read())
			msg.exponent = msg.certificate.public_key.native["public_key"]["public_exponent"]
			msg.modulus = msg.certificate.public_key.native["public_key"]["modulus"]
			msg.bitlen = msg.certificate.public_key.native["public_key"]["modulus"].bit_length()

		return msg
	
	def encrypt(self, secret: bytes):
		# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-rdpbcgr/761e2583-6406-4a71-bfec-cca52294c099
		# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-rdpbcgr/6fd7c8aa-884b-4b43-b036-c86d9d6737e6 <<< little???!!!
		
		temp = pow(int.from_bytes(secret, byteorder='little', signed = False), self.exponent, self.modulus)
		return temp.to_bytes((temp.bit_length() + 7) // 8, 'little').ljust((self.bitlen // 8) + 8, b'\x00')


	def __repr__(self):
		t = '==== SERVER_CERTIFICATE ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t

# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-rdpbcgr/fe93545c-772a-4ade-9d02-ad1e0d81b6af
class RSA_PUBLIC_KEY:
	def __init__(self):
		self.magic:bytes = b'\x31\x41\x53\x52'
		self.keylen:int = None
		self.bitlen:int = None
		self.datalen:int = None
		self.pubExp:int = None
		self.modulus:bytes = None
		
	def to_bytes(self):
		t = self.magic
		t += len(self.modulus).to_bytes(4, byteorder='little', signed = False)
		t += self.bitlen.to_bytes(4, byteorder='little', signed = False)
		t += self.datalen.to_bytes(4, byteorder='little', signed = False)
		t += self.pubExp.to_bytes(4, byteorder='little', signed = False)
		t += self.modulus
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return RSA_PUBLIC_KEY.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = RSA_PUBLIC_KEY()
		msg.magic = buff.read(4)
		msg.keylen = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		msg.bitlen = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		msg.datalen = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		msg.pubExp = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		msg.modulus = buff.read(msg.keylen)
		return msg

	def __repr__(self):
		t = '==== RSA_PUBLIC_KEY ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t


# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-rdpbcgr/a37d449a-73ac-4f00-9b9d-56cefc954634
class PROPRIETARYSERVERCERTIFICATE:
	def __init__(self):
		self.dwSigAlgId:int = None
		self.dwKeyAlgId:int = None
		self.wPublicKeyBlobType:int = None
		self.wPublicKeyBlobLen:int = None
		self.PublicKeyBlob:RSA_PUBLIC_KEY = None
		self.wSignatureBlobType:int = None
		self.wSignatureBlobLen:int = None
		self.SignatureBlob:bytes = None
		
	def to_bytes(self):
		t = self.dwSigAlgId.to_bytes(4, byteorder='little', signed = False)
		t += self.dwKeyAlgId.to_bytes(4, byteorder='little', signed = False)
		t += self.wPublicKeyBlobType.to_bytes(2, byteorder='little', signed = False)
		t += len(self.PublicKeyBlob.to_bytes()).to_bytes(2, byteorder='little', signed = False)
		t += self.PublicKeyBlob.to_bytes()
		t += self.wSignatureBlobType.to_bytes(2, byteorder='little', signed = False)
		t += len(self.SignatureBlob).to_bytes(2, byteorder='little', signed = False)
		t += self.SignatureBlob
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return PROPRIETARYSERVERCERTIFICATE.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = PROPRIETARYSERVERCERTIFICATE()
		msg.dwSigAlgId = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		msg.dwKeyAlgId = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		msg.wPublicKeyBlobType = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.wPublicKeyBlobLen = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.PublicKeyBlob = RSA_PUBLIC_KEY.from_bytes(buff.read(msg.wPublicKeyBlobLen-8))
		zeros = buff.read(8) #
		msg.wSignatureBlobType = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.wSignatureBlobLen = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.SignatureBlob = buff.read(msg.wSignatureBlobLen-8)
		zeros = buff.read(8) #
		return msg

	def __repr__(self):
		t = '==== PROPRIETARYSERVERCERTIFICATE ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t


# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-rdpbcgr/379a020e-9925-4b4f-98f3-7d634e10b411
class TS_UD_SC_SEC1:
	def __init__(self):
		self.length:int = None
		self.type:TS_UD_TYPE = TS_UD_TYPE.SC_SECURITY
		self.encryptionMethod:ENCRYPTION_FLAG = None
		self.encryptionLevel:ENCRYPTION_LEVEL = None
		self.serverRandomLen:int = None
		self.serverCertLen:int = None
		self.serverRandom:bytes = None
		self.serverCertificate:SERVER_CERTIFICATE = None
		
	def to_bytes(self):
		def finish(t):
			t = (len(t)+4).to_bytes(2, byteorder='little', signed = False) + t
			t = self.type.value.to_bytes(2, byteorder='little', signed = False) + t
			return t
		t = self.encryptionMethod.value.to_bytes(4, byteorder='little', signed = False)
		t += self.encryptionLevel.value.to_bytes(4, byteorder='little', signed = False)
		t += len(self.serverRandom).to_bytes(4, byteorder='little', signed = False)
		t += len(self.serverCertificate).to_bytes(4, byteorder='little', signed = False)
		t += self.serverRandom
		if self.serverCertificate is not None:
			t += self.serverCertificate.to_bytes()
		return finish(t)

	@staticmethod
	def from_bytes(bbuff: bytes):
		return TS_UD_SC_SEC1.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = TS_UD_SC_SEC1()
		msg.type = TS_UD_TYPE(int.from_bytes(buff.read(2), byteorder='little', signed = False))
		msg.length = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.encryptionMethod = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		msg.encryptionLevel = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		msg.serverRandomLen = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		msg.serverCertLen = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		msg.serverRandom = buff.read(msg.serverRandomLen)

		if msg.serverCertLen > 0:
			msg.serverCertificate = SERVER_CERTIFICATE.from_bytes(buff.read(msg.serverCertLen))
		
		return msg

	def __repr__(self):
		t = '==== TS_UD_SC_SEC1 ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t



import asyncio
import traceback

from aardwolf.protocol.T124.userdata.constants import ChannelOption

class Channel:
	def __init__(self, name:str, options:ChannelOption = ChannelOption.INITIALIZED|ChannelOption.ENCRYPT_RDP):
		self.name = name
		self.options = options
		self.channel_id = None
		self.connection = None

	def get_name(self):
		return self.name
	
	async def disconnect(self):
		try:
			_, err = await self.stop()
			if err is not None:
				raise err
		except:
			pass

	async def run(self, connection):
		try:
			self.connection = connection
			_, err = await self.start()
			if err is not None:
				raise err
			return True, None
		except Exception as e:
			traceback.print_exc()
			return None, e


	async def start(self):
		# Override this when implementing new channel type
		try:
			return True, None
		except Exception as e:
			return None, e
	
	async def stop(self):
		# Override this when implementing new channel type
		try:
			return True, None
		except Exception as e:
			return None, e

	async def process_channel_data(self, data):
		# This function will be called when channel data is recieved from the server
		raise NotImplementedError()
	
	async def process_user_data(self, data):
		# This function will be called when the channel recieves a new user data
		# Override this function to implement user data processing
		raise NotImplementedError()
	
	async def send_channel_data(self, dataobj, sec_hdr, datacontrol_hdr, sharecontrol_hdr, is_fastpath):
		await self.connection.handle_out_data(dataobj, sec_hdr, datacontrol_hdr, sharecontrol_hdr, self.channel_id, is_fastpath)

	async def send_user_data(self, msg):
		await self.connection.ext_out_queue.put(msg)


class MCSChannel(Channel):
	name = 'MCS'
	def __init__(self, iosettings = None):
		Channel.__init__(self, self.name, ChannelOption.INITIALIZED|ChannelOption.ENCRYPT_RDP)
		self.out_queue = asyncio.Queue()
	

	async def process_channel_data(self, data):
		# This function will be called when channel data is recieved from the server
		#print('Channel data in! "%s(%s)" <- %s' % (self.name, self.channel_id, data))
		await self.out_queue.put((data, None))



import io
import enum
from aardwolf.protocol.T124.userdata.constants import TS_UD_TYPE, MultitransportFlags

# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-rdpbcgr/bf7201d4-9ed9-4dfe-9f6f-f2d68a7367ed
class TS_UD_SC_MULTITRANSPORT:
	def __init__(self):
		self.length:int = None
		self.type:TS_UD_TYPE = TS_UD_TYPE.SC_MULTITRANSPORT
		self.flags:MultitransportFlags = None
		
	def to_bytes(self):
		def finish(t):
			t = (len(t)+4).to_bytes(2, byteorder='little', signed = False) + t
			t = self.type.value.to_bytes(2, byteorder='little', signed = False) + t
			return t
		t = self.flags.to_bytes(4, byteorder='little', signed = False)
		return finish(t)

	@staticmethod
	def from_bytes(bbuff: bytes):
		return TS_UD_SC_MULTITRANSPORT.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = TS_UD_SC_MULTITRANSPORT()
		msg.type = TS_UD_TYPE(int.from_bytes(buff.read(2), byteorder='little', signed = False))
		msg.length = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.flags = MultitransportFlags(int.from_bytes(buff.read(4), byteorder='little', signed = False))
		return msg

	def __repr__(self):
		t = '==== TS_UD_SC_MULTITRANSPORT ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t


import enum

# https://github.com/serisman/FreeRDP/blob/master/include/freerdp/constants/pdu.h

class TPDU_TYPE(enum.Enum):
	CONNECTION_REQUEST = 0xE
	CONNECTION_CONFIRM = 0xD
	DISCONNECT_REQUEST = 0x8
	DATA = 0xF
	ERROR = 0x7

class NEGOTIATIONType(enum.Enum):
	NEG_REQ = 0x1
	NEG_RSP = 0x2
	NEG_FAILURE = 0x3

class NEG_FLAGS(enum.IntFlag):
	RESTRICTED_ADMIN_MODE_REQUIRED = 0x01 #Indicates that the client requires credential-less logon over CredSSP (also known as "restricted admin mode"). If the server supports this mode then it is acceptable for the client to send empty credentials in the TSPasswordCreds structure defined in [MS-CSSP] section 2.2.1.2.1.<2>
	REDIRECTED_AUTHENTICATION_MODE_REQUIRED = 0x02 #Indicates that the client requires credential-less logon over CredSSP with redirected authentication over CredSSP (also known as "Remote Credential Guard"). If the server supports this mode, the client can send a redirected logon buffer in the TSRemoteGuardCreds structure defined in [MS-CSSP] section 2.2.1.2.3.
	CORRELATION_INFO_PRESENT = 0x08 #The optional rdpCorrelationInfo field of the 224 Connection Request PDU (section 2.2.1.1) is present.

class RESP_FLAGS(enum.IntFlag):
	EXTENDED_CLIENT_DATA_SUPPORTED = 0x01 #The server supports Extended Client Data Blocks in the GCC Conference Create Request user data (section 2.2.1.3).
	DYNVC_GFX_PROTOCOL_SUPPORTED = 0x02 #The server supports the Graphics Pipeline Extension Protocol described in [MS-RDPEGFX] sections 1, 2, and 3.
	NEGRSP_FLAG_RESERVED = 0x04 #An unused flag that is reserved for future use. This flag SHOULD be ignored by the client.
	RESTRICTED_ADMIN_MODE_SUPPORTED = 0x08 #Indicates that the server supports credential-less logon over CredSSP (also known as "restricted admin mode") and it is acceptable for the client to send empty credentials in the TSPasswordCreds structure defined in [MS-CSSP] section 2.2.1.2.1.<3>
	REDIRECTED_AUTHENTICATION_MODE_SUPPORTED = 0x10 #Indicates that the server supports credential-less logon over CredSSP with credential redirection (also known as "Remote Credential Guard"). The client can send a redirected logon buffer in the TSRemoteGuardCreds structure defined in [MS-CSSP] section 2.2.1.2.3.

class SUPP_PROTOCOLS(enum.IntFlag):
	RDP = 0x00000000
	SSL = 0x00000001
	HYBRID = 0x00000002
	RDSTLS = 0x00000004
	HYBRID_EX = 0x00000008

class FAIL_CODE(enum.Enum):
	SSL_REQUIRED_BY_SERVER = 0x00000001 #The server requires that the client support Enhanced RDP Security (section 5.4) with either TLS 1.0, 1.1 or 1.2 (section 5.4.5.1) or CredSSP (section 5.4.5.2). If only CredSSP was requested then the server only supports TLS.
	SSL_NOT_ALLOWED_BY_SERVER = 0x00000002 #The server is configured to only use Standard RDP Security mechanisms (section 5.3) and does not support any External Security Protocols (section 5.4.5).
	SSL_CERT_NOT_ON_SERVER = 0x00000003 #The server does not possess a valid authentication certificate and cannot initialize the External Security Protocol Provider (section 5.4.5).
	INCONSISTENT_FLAGS = 0x00000004 #The list of requested security protocols is not consistent with the current security protocol in effect. This error is only possible when the Direct Approach (sections 5.4.2.2 and 1.3.1.2) is used and an External Security Protocol (section 5.4.5) is already being used.
	HYBRID_REQUIRED_BY_SERVER = 0x00000005 #The server requires that the client support Enhanced RDP Security (section 5.4) with CredSSP (section 5.4.5.2).
	SSL_WITH_USER_AUTH_REQUIRED_BY_SERVER = 0x00000006 #The server requires that the client support Enhanced RDP Security (section 5.4) with TLS 1.0, 1.1 or 1.2 (section 5.4.5.1) and certificate-based client authentication.<4>

class ERRORINFO(enum.Enum):
	# Protocol-independent codes: */
	RPC_INITIATED_DISCONNECT = 0x0001
	RPC_INITIATED_LOGOFF = 0x0002
	IDLE_TIMEOUT = 0x0003
	LOGON_TIMEOUT = 0x0004
	DISCONNECTED_BY_OTHERCONNECTION = 0x0005
	OUT_OF_MEMORY = 0x0006
	SERVER_DENIED_CONNECTION = 0x0007
	SERVER_INSUFFICIENT_PRIVILEGES = 0x0009
	SERVER_FRESH_CREDENTIALS_REQUIRED = 0x000A
	RPC_INITIATED_DISCONNECT_BYUSER = 0x000B
	# Protocol-independent licensing codes: */
	LICENSE_INTERNAL = 0x0100
	LICENSE_NO_LICENSE_SERVER = 0x0101
	LICENSE_NO_LICENSE = 0x0102
	LICENSE_BAD_CLIENT_MSG = 0x0103
	LICENSE_HWID_DOESNT_MATCH_LICENSE = 0x0104
	LICENSE_BAD_CLIENT_LICENSE = 0x0105
	LICENSE_CANT_FINISH_PROTOCOL = 0x0106
	LICENSE_CLIENT_ENDED_PROTOCOL = 0x0107
	LICENSE_BAD_CLIENT_ENCRYPTION = 0x0108
	LICENSE_CANT_UPGRADE_LICENSE = 0x0109
	LICENSE_NO_REMOTE_CONNECTIONS = 0x010A
	# RDP specific codes: */
	UNKNOWNPDUTYPE2 = 0x10C9
	UNKNOWNPDUTYPE = 0x10CA
	DATAPDUSEQUENCE = 0x10CB
	CONTROLPDUSEQUENCE = 0x10CD
	INVALIDCONTROLPDUACTION = 0x10CE
	INVALIDINPUTPDUTYPE = 0x10CF
	INVALIDINPUTPDUMOUSE = 0x10D0
	INVALIDREFRESHRECTPDU = 0x10D1
	CREATEUSERDATAFAILED = 0x10D2
	CONNECTFAILED = 0x10D3
	CONFIRMACTIVEWRONGSHAREID = 0x10D4
	CONFIRMACTIVEWRONGORIGINATOR = 0x10D5
	PERSISTENTKEYPDUBADLENGTH = 0x10DA
	PERSISTENTKEYPDUILLEGALFIRST = 0x10DB
	PERSISTENTKEYPDUTOOMANYTOTALKEYS = 0x10DC
	PERSISTENTKEYPDUTOOMANYCACHEKEYS = 0x10DD
	INPUTPDUBADLENGTH = 0x10DE
	BITMAPCACHEERRORPDUBADLENGTH = 0x10DF
	SECURITYDATATOOSHORT = 0x10E0
	VCHANNELDATATOOSHORT = 0x10E1
	SHAREDATATOOSHORT = 0x10E2
	BADSUPRESSOUTPUTPDU = 0x10E3
	CONFIRMACTIVEPDUTOOSHORT = 0x10E5
	CAPABILITYSETTOOSMALL = 0x10E7
	CAPABILITYSETTOOLARGE = 0x10E8
	NOCURSORCACHE = 0x10E9
	BADCAPABILITIES = 0x10EA
	VIRTUALCHANNELDECOMPRESSIONERR = 0x10EC
	INVALIDVCCOMPRESSIONTYPE = 0x10ED
	INVALIDCHANNELID = 0x10EF
	VCHANNELSTOOMANY = 0x10F0
	REMOTEAPPSNOTENABLED = 0x10F3
	CACHECAPNOTSET = 0x10F4
	BITMAPCACHEERRORPDUBADLENGTH2 = 0x10F5
	OFFSCRCACHEERRORPDUBADLENGTH = 0x10F6
	DNGCACHEERRORPDUBADLENGTH = 0x10F7
	GDIPLUSPDUBADLENGTH = 0x10F8
	SECURITYDATATOOSHORT2 = 0x1111
	SECURITYDATATOOSHORT3 = 0x1112
	SECURITYDATATOOSHORT4 = 0x1113
	SECURITYDATATOOSHORT5 = 0x1114
	SECURITYDATATOOSHORT6 = 0x1115
	SECURITYDATATOOSHORT7 = 0x1116
	SECURITYDATATOOSHORT8 = 0x1117
	SECURITYDATATOOSHORT9 = 0x1118
	SECURITYDATATOOSHORT10 = 0x1119
	SECURITYDATATOOSHORT11 = 0x111A
	SECURITYDATATOOSHORT12 = 0x111B
	SECURITYDATATOOSHORT13 = 0x111C
	SECURITYDATATOOSHORT14 = 0x111D
	SECURITYDATATOOSHORT15 = 0x111E
	SECURITYDATATOOSHORT16 = 0x111F
	SECURITYDATATOOSHORT17 = 0x1120
	SECURITYDATATOOSHORT18 = 0x1121
	SECURITYDATATOOSHORT19 = 0x1122
	SECURITYDATATOOSHORT20 = 0x1123
	SECURITYDATATOOSHORT21 = 0x1124
	SECURITYDATATOOSHORT22 = 0x1125
	SECURITYDATATOOSHORT23 = 0x1126
	BADMONITORDATA = 0x1129
	UPDATESESSIONKEYFAILED = 0x1191
	DECRYPTFAILED = 0x1192
	ENCRYPTFAILED = 0x1193
	ENCPKGMISMATCH = 0x1194
	DECRYPTFAILED2 = 0x1195




import enum
import io
from typing import List
from aardwolf.protocol.T128.share import TS_SHARECONTROLHEADER
from aardwolf.protocol.pdu.capabilities import TS_CAPS_SET

class TS_DEMAND_ACTIVE_PDU:
	def __init__(self):
		self.shareControlHeader:TS_SHARECONTROLHEADER = None
		self.shareID:int = None
		self.lengthSourceDescriptor:int = None
		self.lengthCombinedCapabilities:int = None
		self.sourceDescriptor:bytes = None #cli: MTSC\0x00 srv: RDP\x00
		self.numberCapabilities:int = None
		self.pad2Octets:int = None
		self.capabilitySets:List[TS_CAPS_SET] = []
		self.sessionId:int = None

	def to_bytes(self):
		capdata = b''
		for cap in self.capabilitySets:
			capdata += cap.to_bytes()
		
		t = self.shareID.to_bytes(4, byteorder='little', signed = False)
		t += len(self.sourceDescriptor).to_bytes(2, byteorder='little', signed = False)
		t += (len(capdata) + 2 + 2).to_bytes(2, byteorder='little', signed = False)
		t += self.sourceDescriptor
		t += len(self.capabilitySets).to_bytes(2, byteorder='little', signed = False)
		t += self.pad2Octets.to_bytes(2, byteorder='little', signed = False)
		t += capdata
		t += self.sessionId.to_bytes(4, byteorder='little', signed = False)
		#self.shareControlHeader.totalLength = 6 + len(t)
		#t = self.shareControlHeader.to_bytes() + t
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return TS_DEMAND_ACTIVE_PDU.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = TS_DEMAND_ACTIVE_PDU()
		msg.shareControlHeader = TS_SHARECONTROLHEADER.from_buffer(buff)
		msg.shareID = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		msg.lengthSourceDescriptor = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.lengthCombinedCapabilities = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.sourceDescriptor = buff.read(msg.lengthSourceDescriptor)
		msg.numberCapabilities = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.pad2Octets = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		for _ in range(msg.numberCapabilities):
			msg.capabilitySets.append(TS_CAPS_SET.from_buffer(buff))
		msg.sessionId = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		
		
		return msg

	def __repr__(self):
		t = '==== TS_DEMAND_ACTIVE_PDU ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t




import io

# https://datatracker.ietf.org/doc/html/rfc2126#section-4.3
class TPKT:
	def __init__(self):
		self.version = 3
		self.reserved = 0
		self.length = None
		self.tpdu = None
		
	def to_bytes(self):
		t  = self.version.to_bytes(1, byteorder='big', signed = False)
		t += self.reserved.to_bytes(1, byteorder='big', signed = False)
		t += (len(self.tpdu)+4).to_bytes(2, byteorder='big', signed = False)
		t += self.tpdu
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return TPKT.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = TPKT()
		msg.version = int.from_bytes(buff.read(1), byteorder='big', signed=False)
		msg.reserved = int.from_bytes(buff.read(1), byteorder='big', signed=False)
		msg.length = int.from_bytes(buff.read(2), byteorder='big', signed=False)
		msg.tpdu = buff.read(msg.length-4)
		return msg

	def __repr__(self):
		t = '==== TPKT ====\r\n'
		t += 'version %s\r\n' % self.version
		t += 'reserved %s\r\n' % self.reserved
		t += 'length %s\r\n' % self.length
		t += 'tpdu %s\r\n' % self.tpdu
		return t


import io
import enum

class GLYPH_SUPPORT(enum.Enum):
	NONE = 0x0000 #The client does not support glyph caching. All text output will be sent to the client as expensive Bitmap Updates (sections 2.2.9.1.1.3.1.2 and 2.2.9.1.2.1.2).
	PARTIAL = 0x0001 #Indicates support for Revision 1 Cache Glyph Secondary Drawing Orders ([MS-RDPEGDI] section 2.2.2.2.1.2.5).
	FULL = 0x0002 #Indicates support for Revision 1 Cache Glyph Secondary Drawing Orders ([MS-RDPEGDI] section 2.2.2.2.1.2.5).
	ENCODE = 0x0003 #Indicates support for Revision 2 Cache Glyph Secondary Drawing Orders ([MS-RDPEGDI] section 2.2.2.2.1.2.6).

class TS_GLYPHCACHE_CAPABILITYSET:
	def __init__(self):
		self.GlyphCache:bytes = b'\x00'*40 #https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-rdpbcgr/cae26830-263c-4c1e-97c2-b561faded3d9
		self.FragCache:bytes = b'\x00'*4
		self.GlyphSupportLevel:GLYPH_SUPPORT = GLYPH_SUPPORT.NONE
		self.pad2octets: bytes = b'\x00'*2

	def to_bytes(self):
		t = self.GlyphCache
		t += self.FragCache
		t += self.GlyphSupportLevel.value.to_bytes(2, byteorder='little', signed = False)
		t += self.pad2octets
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return TS_GLYPHCACHE_CAPABILITYSET.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = TS_GLYPHCACHE_CAPABILITYSET()
		msg.GlyphCache = buff.read(40)
		msg.FragCache = buff.read(4)
		msg.GlyphSupportLevel = GLYPH_SUPPORT(int.from_bytes(buff.read(2), byteorder='little', signed = False))
		msg.pad2octets = buff.read(2)
		return msg

	def __repr__(self):
		t = '==== TS_GLYPHCACHE_CAPABILITYSET ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t


import io
import enum

class BRUSH_FLAG(enum.Enum):
	DEFAULT = 0x00000000 #Support for solid-color and monochrome pattern brushes with no caching. This is an RDP 4.0 implementation.
	COLOR_8x8 = 0x00000001 #Ability to handle color brushes (4-bit or 8-bit in RDP 5.0; 4-bit, 8-bit, 16-bit, or 24-bit in all other RDP versions, except for RDP 4.0) and caching. Brushes are limited to 8-by-8 pixels.
	COLOR_FULL = 0x00000002 #Ability to handle color brushes (4-bit or 8-bit in RDP 5.0; 4-bit, 8-bit, 16-bit, or 24-bit in all other RDP versions, except for RDP 4.0) and caching. Brushes can have arbitrary dimensions.

class TS_BRUSH_CAPABILITYSET:
	def __init__(self):
		self.brushSupportLevel:BRUSH_FLAG = BRUSH_FLAG.DEFAULT #FONTSUPPORT_FONTLIST

	def to_bytes(self):
		t = self.brushSupportLevel.value.to_bytes(4, byteorder='little', signed = False)
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return TS_BRUSH_CAPABILITYSET.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = TS_BRUSH_CAPABILITYSET()
		msg.brushSupportLevel = BRUSH_FLAG(int.from_bytes(buff.read(4), byteorder='little', signed = False))
		return msg

	def __repr__(self):
		t = '==== TS_BRUSH_CAPABILITYSET ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t


import io
import enum
from aardwolf.protocol.pdu.input.keyboard import KBDFLAGS

class TS_UNICODE_KEYBOARD_EVENT:
	def __init__(self):
		self.keyboardFlags:KBDFLAGS = None #only release flag is allowed here
		self.unicodeCode:int = None #maybe string? #The Unicode character input code.
		self.pad2Octets:bytes = b'\x00\x00'

	def to_bytes(self):
		t = self.keyboardFlags.to_bytes(2, byteorder='little', signed=False)
		t += self.unicodeCode[0].encode('utf-16-le')
		t += self.pad2Octets
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return TS_UNICODE_KEYBOARD_EVENT.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = TS_UNICODE_KEYBOARD_EVENT()
		msg.keyboardFlags = KBDFLAGS(int.from_bytes(buff.read(2), byteorder='little', signed=False))
		msg.unicodeCode = buff.read(2).decode('utf-16-le')
		msg.pad2Octets = buff.read(2)
		return msg

	def __repr__(self):
		t = '==== TS_UNICODE_KEYBOARD_EVENT ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t


import enum
import io


class TS_SYSTEMTIME:
	def __init__(self):
		self.wYear:int = 0
		self.wMonth:int = None
		self.wDayOfWeek:int = None
		self.wDay:int = None
		self.wHour:int = None
		self.wMinute:int = None
		self.wSecond:int = 0
		self.wMilliseconds:int = 0

	def to_bytes(self):
		t = self.wYear.to_bytes(2, byteorder='little', signed = False)
		t += self.wMonth.to_bytes(2, byteorder='little', signed = False)
		t += self.wDayOfWeek.to_bytes(2, byteorder='little', signed = False)
		t += self.wDay.to_bytes(2, byteorder='little', signed = False)
		t += self.wHour.to_bytes(2, byteorder='little', signed = False)
		t += self.wMinute.to_bytes(2, byteorder='little', signed = False)
		t += self.wSecond.to_bytes(2, byteorder='little', signed = False)
		t += self.wMilliseconds.to_bytes(2, byteorder='little', signed = False)
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return TS_SYSTEMTIME.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = TS_SYSTEMTIME()
		msg.wYear = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.wMonth = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.wDayOfWeek = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.wDay = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.wHour = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.wMinute = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.wSecond = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.wMilliseconds = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		return msg

	def __repr__(self):
		t = '==== TS_SYSTEMTIME ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t

class TS_TIME_ZONE_INFORMATION:
	def __init__(self):
		self.Bias:int = None
		self.StandardName:bytes = None
		self.StandardDate:TS_SYSTEMTIME = None
		self.StandardBias:int = None
		self.DaylightName:bytes = None
		self.DaylightDate:TS_SYSTEMTIME = None
		self.DaylightBias:int = 0

	def to_bytes(self):
		t = self.Bias.to_bytes(4, byteorder='little', signed = False)
		t += self.StandardName
		t += self.StandardDate.to_bytes()
		t += self.StandardBias.to_bytes(4, byteorder='little', signed = False)
		t += self.DaylightName
		t += self.DaylightDate.to_bytes()
		t += self.DaylightBias.to_bytes(4, byteorder='little', signed = False)
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return TS_TIME_ZONE_INFORMATION.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = TS_TIME_ZONE_INFORMATION()
		msg.Bias = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		msg.StandardName = buff.read(64)
		msg.StandardDate = TS_SYSTEMTIME.from_buffer(buff)
		msg.StandardBias = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		msg.DaylightName = buff.read(64)
		msg.DaylightDate = TS_SYSTEMTIME.from_buffer(buff)
		msg.DaylightBias = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		return msg

	def __repr__(self):
		t = '==== TS_TIME_ZONE_INFORMATION ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t

class PERF(enum.IntFlag):
	DISABLE_WALLPAPER = 0x00000001 #Disable desktop wallpaper.
	DISABLE_FULLWINDOWDRAG = 0x00000002 #Disable full-window drag (only the window outline is displayed when the window is moved).
	DISABLE_MENUANIMATIONS = 0x00000004 #Disable menu animations.
	DISABLE_THEMING = 0x00000008 #Disable user interface themes.
	RESERVED1 = 0x00000010 #An unused flag that is reserved for future use. This flag SHOULD be ignored by the server.
	DISABLE_CURSOR_SHADOW = 0x00000020 #Disable mouse cursor shadows.
	DISABLE_CURSORSETTINGS = 0x00000040 #Disable cursor blinking.
	ENABLE_FONT_SMOOTHING = 0x00000080 #Enable font smoothing.<18>
	ENABLE_DESKTOP_COMPOSITION = 0x00000100 #Enable Desktop Composition ([MS-RDPEDC] sections 1, 2 and 3; and [MS-RDPCR2] sections 1, 2 and 3). The usage of Desktop Composition in a remote session requires that the color depth be 32 bits per pixel (bpp). (See the description of the highColorDepth, supportedColorDepths and earlyCapabilityFlags (specifically the RNS_UD_CS_WANT_32BPP_SESSION (0x0002) flag) fields in section 2.2.1.3.2 for background on setting the remote session color depth to 32 bpp.)<19>
	RESERVED2 = 0x80000000 #An unused flag that is reserved for future use. This flag SHOULD be ignored by the server.

class CLI_AF(enum.Enum):
	AF_INET  = 0x0002 #The clientAddress field contains an IPv4 address.
	AF_INET6 = 0x0017 #The clientAddress field contains an IPv6 address.

# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-rdpbcgr/732394f5-e2b5-4ac5-8a0a-35345386b0d1
class TS_EXTENDED_INFO_PACKET:
	def __init__(self):
		self.clientAddressFamily:CLI_AF = None
		self.cbClientAddress:int = None
		self.clientAddress:str = None
		self.cbClientDir:int = None
		self.clientDir:str = None
		self.clientTimeZone:TS_TIME_ZONE_INFORMATION = None
		self.clientSessionId:int = 0
		self.performanceFlags:PERF = None
		self.cbAutoReconnectCookie:int = None
		self.autoReconnectCookie:bytes = None
		self.reserved1:bytes = b'\x00\x00'
		self.reserved2:bytes = b'\x00\x00'
		self.cbDynamicDSTTimeZoneKeyName:int = None
		self.dynamicDSTTimeZoneKeyName:str = None
		self.dynamicDaylightTimeDisabled:bool = None

	def to_bytes(self):
		t = self.clientAddressFamily.value.to_bytes(2, byteorder='little', signed = False)
		t += len((self.clientAddress+'\x00').encode('utf-16-le')).to_bytes(2, byteorder='little', signed = False)
		t += (self.clientAddress+'\x00').encode('utf-16-le')
		t += len((self.clientDir+'\x00').encode('utf-16-le')).to_bytes(2, byteorder='little', signed = False)
		t += (self.clientDir+'\x00').encode('utf-16-le')
		if self.clientTimeZone is None:
			return t
		t += self.clientTimeZone.to_bytes()
		if self.clientSessionId is None:
			return t
		t += self.clientSessionId.to_bytes(4, byteorder='little', signed = False)
		if self.performanceFlags is None:
			return t
		t += self.performanceFlags.to_bytes(4, byteorder='little', signed = False)
		if self.autoReconnectCookie is None:
			return t
		t += len(self.autoReconnectCookie).to_bytes(2, byteorder='little', signed = False)
		if self.autoReconnectCookie is None:
			return t
		t += self.autoReconnectCookie
		if self.reserved1 is None:
			return t
		t += self.reserved1
		if self.reserved2 is None:
			return t
		t += self.reserved2
		if self.dynamicDSTTimeZoneKeyName is None:
			return t
		t += len(self.dynamicDSTTimeZoneKeyName).to_bytes(2, byteorder='little', signed = False)
		if self.dynamicDSTTimeZoneKeyName is None:
			return t
		t += self.dynamicDSTTimeZoneKeyName.encode()
		if self.dynamicDaylightTimeDisabled is None:
			return t
		t += int(self.dynamicDaylightTimeDisabled).to_bytes(2, byteorder='little', signed = False)
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return TS_EXTENDED_INFO_PACKET.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = TS_EXTENDED_INFO_PACKET()		
		msg.clientAddressFamily = CLI_AF(int.from_bytes(buff.read(2), byteorder='little', signed= False))
		msg.cbClientAddress = int.from_bytes(buff.read(2), byteorder='little', signed= False)
		msg.clientAddress = buff.read(msg.cbClientAddress).decode('utf-16-le').replace('\x00', '')
		msg.cbClientDir = int.from_bytes(buff.read(2), byteorder='little', signed= False)
		msg.clientDir = buff.read(msg.cbClientDir).decode('utf-16-le').replace('\x00', '')
		if buff.read(1) == b'':
			return msg
		buff.seek(-1,1)
		msg.clientTimeZone = TS_TIME_ZONE_INFORMATION.from_buffer(buff)
		if buff.read(1) == b'':
			return msg
		buff.seek(-1,1)
		msg.clientSessionId = int.from_bytes(buff.read(4), byteorder='little', signed= False)
		if buff.read(1) == b'':
			return msg
		buff.seek(-1,1)
		msg.performanceFlags = PERF(int.from_bytes(buff.read(4), byteorder='little', signed= False))
		if buff.read(1) == b'':
			return msg
		buff.seek(-1,1)
		msg.cbAutoReconnectCookie = int.from_bytes(buff.read(2), byteorder='little', signed= False)
		if buff.read(1) == b'':
			return msg
		buff.seek(-1,1)
		msg.autoReconnectCookie = buff.read(28) #always 28 docu
		if buff.read(1) == b'':
			return msg
		buff.seek(-1,1)
		msg.reserved1 = buff.read(2)
		if buff.read(1) == b'':
			return msg
		buff.seek(-1,1)
		msg.reserved2 = buff.read(2)
		if buff.read(1) == b'':
			return msg
		buff.seek(-1,1)
		msg.cbDynamicDSTTimeZoneKeyName = int.from_bytes(buff.read(2), byteorder='little', signed= False)
		if buff.read(1) == b'':
			return msg
		buff.seek(-1,1)
		msg.dynamicDSTTimeZoneKeyName = buff.read(msg.cbDynamicDSTTimeZoneKeyName).decode()
		if buff.read(1) == b'':
			return msg
		buff.seek(-1,1)
		msg.dynamicDaylightTimeDisabled = bool(int.from_bytes(buff.read(2), byteorder='little', signed= False))
		return msg

	def __repr__(self):
		t = '==== TS_EXTENDED_INFO_PACKET ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t


import io
import enum


class TS_CONTROL_CAPABILITYSET:
	def __init__(self):
		self.controlFlags:int = 0
		self.remoteDetachFlag:bool = False
		self.controlInterest:int = 2
		self.detachInterest:int = 2

	def to_bytes(self):
		t = self.controlFlags.to_bytes(2, byteorder='little', signed = False)
		t += int(self.remoteDetachFlag).to_bytes(2, byteorder='little', signed = False)
		t += self.controlInterest.to_bytes(2, byteorder='little', signed = False)
		t += self.detachInterest.to_bytes(2, byteorder='little', signed = False)
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return TS_CONTROL_CAPABILITYSET.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = TS_CONTROL_CAPABILITYSET()
		msg.controlFlags = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.remoteDetachFlag = bool(int.from_bytes(buff.read(2), byteorder='little', signed = False))
		msg.controlInterest = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.detachInterest = int.from_bytes(buff.read(2), byteorder='little', signed = False)	
		return msg

	def __repr__(self):
		t = '==== TS_CONTROL_CAPABILITYSET ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t



import io
import enum
from aardwolf.protocol.T124.userdata.constants import TS_UD_TYPE, COLOR_DEPTH, HIGH_COLOR_DEPTH, \
	SUPPORTED_COLOR_DEPTH, CONNECTION_TYPE, RNS_UD_CS, ORIENTATION
from aardwolf.protocol.x224.constants import SUPP_PROTOCOLS

# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-rdpbcgr/8a36630c-9c8e-4864-9382-2ec9d6f368ca
class TS_UD_CS_CORE:
	def __init__(self):
		self.type:TS_UD_TYPE = TS_UD_TYPE.CS_CORE
		self.length = None

		self.version: int = 0x00080005
		self.desktopWidth: int = None
		self.desktopHeight: int = None
		self.colorDepth: COLOR_DEPTH = None
		self.SASSequence: int = 0xAA03
		self.keyboardLayout: int = 1033
		self.clientBuild: int = None
		self.clientName: str = None
		self.keyboardType: int = 0x00000004
		self.keyboardSubType: int = 0
		self.keyboardFunctionKey: int = 12
		self.imeFileName: str = None
		self.postBeta2ColorDepth: COLOR_DEPTH = None
		self.clientProductId:int = None
		self.serialNumber:int = None
		self.highColorDepth: HIGH_COLOR_DEPTH = None
		self.supportedColorDepths: SUPPORTED_COLOR_DEPTH = None
		self.earlyCapabilityFlags:RNS_UD_CS = None
		self.clientDigProductId:bytes = None
		self.connectionType:CONNECTION_TYPE = None
		self.pad1octet:bytes = None
		self.serverSelectedProtocol:SUPP_PROTOCOLS = None
		self.desktopPhysicalWidth:int = None
		self.desktopPhysicalHeight:int = None
		self.desktopOrientation:ORIENTATION = None
		self.desktopScaleFactor:int = None
		self.deviceScaleFactor:int = None

		
	def to_bytes(self):
		def finish(t):
			t = (len(t)+4).to_bytes(2, byteorder='little', signed = False) + t
			t = self.type.value.to_bytes(2, byteorder='little', signed = False) + t
			return t
		clientName = self.clientName.encode('utf-16-le').ljust(32, b'\x00')
		imeFileName = self.imeFileName.encode('utf-16-le').ljust(64, b'\x00')
		t = self.version.to_bytes(4, byteorder='little', signed = False)
		t += self.desktopWidth.to_bytes(2, byteorder='little', signed = False)
		t += self.desktopHeight.to_bytes(2, byteorder='little', signed = False)
		t += self.colorDepth.value.to_bytes(2, byteorder='little', signed = False)
		t += self.SASSequence.to_bytes(2, byteorder='little', signed = False)
		t += self.keyboardLayout.to_bytes(4, byteorder='little', signed = False)
		t += self.clientBuild.to_bytes(4, byteorder='little', signed = False)
		t += clientName
		t += self.keyboardType.to_bytes(4, byteorder='little', signed = False)
		t += self.keyboardSubType.to_bytes(4, byteorder='little', signed = False)
		t += self.keyboardFunctionKey.to_bytes(4, byteorder='little', signed = False)
		t += imeFileName
		#optional fileds...
		if self.postBeta2ColorDepth is None:
			return finish(t)
		t += self.postBeta2ColorDepth.value.to_bytes(2, byteorder='little', signed = False)
		if self.clientProductId is None:
			return finish(t)
		t += self.clientProductId.to_bytes(2, byteorder='little', signed = False)
		if self.serialNumber is None:
			return finish(t)
		t += self.serialNumber.to_bytes(4, byteorder='little', signed = False)
		if self.highColorDepth is None:
			return finish(t)
		t += self.highColorDepth.value.to_bytes(2, byteorder='little', signed = False)
		if self.supportedColorDepths is None:
			return finish(t)
		t += self.supportedColorDepths.to_bytes(2, byteorder='little', signed = False)
		if self.earlyCapabilityFlags is None:
			return finish(t)
		t += self.earlyCapabilityFlags.to_bytes(2, byteorder='little', signed = False)
		if self.clientDigProductId is None:
			return finish(t)
		t += self.clientDigProductId
		if self.connectionType is None:
			return finish(t)
		t += self.connectionType.value.to_bytes(1, byteorder='little', signed = False)
		if self.pad1octet is None:
			return finish(t)
		t += self.pad1octet
		if self.serverSelectedProtocol is None:
			return finish(t)
		t += self.serverSelectedProtocol.value.to_bytes(4, byteorder='little', signed = False)
		if self.desktopPhysicalWidth is None:
			return finish(t)		
		t += self.desktopPhysicalWidth.to_bytes(4, byteorder='little', signed = False)
		if self.desktopPhysicalHeight is None:
			return finish(t)
		t += self.desktopPhysicalHeight.to_bytes(4, byteorder='little', signed = False)
		if self.desktopOrientation is None:
			return finish(t)
		t += self.desktopOrientation.value.to_bytes(2, byteorder='little', signed = False)
		if self.desktopScaleFactor is None:
			return finish(t)
		t += self.desktopScaleFactor.to_bytes(4, byteorder='little', signed = False)
		if self.deviceScaleFactor is None:
			return finish(t)
		t += self.deviceScaleFactor.to_bytes(4, byteorder='little', signed = False)
		return finish(t)

	@staticmethod
	def from_bytes(bbuff: bytes):
		return TS_UD_CS_CORE.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		def is_end(buff, start, size):
			return buff.tell() - start >= msg.length
		start = buff.tell()
		msg = TS_UD_CS_CORE()
		msg.type = TS_UD_TYPE(int.from_bytes(buff.read(2), byteorder='little', signed = False))
		msg.length = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		dlen = msg.length
		msg.version = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		msg.desktopWidth = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.desktopHeight = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.colorDepth = COLOR_DEPTH(int.from_bytes(buff.read(2), byteorder='little', signed = False))
		msg.SASSequence = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.keyboardLayout = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		msg.clientBuild = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		msg.clientName = buff.read(32).decode('utf-16-le').replace('\x00', '')
		msg.keyboardType = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		msg.keyboardSubType = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		msg.keyboardFunctionKey = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		msg.imeFileName = buff.read(64).decode('utf-16-le').replace('\x00', '')
		if is_end(buff,start, dlen):
			return msg
		msg.postBeta2ColorDepth = COLOR_DEPTH(int.from_bytes(buff.read(2), byteorder='little', signed = False))
		if is_end(buff,start, dlen):
			return msg
		msg.clientProductId = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		if is_end(buff,start, dlen):
			return msg
		msg.serialNumber = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		if is_end(buff,start, dlen):
			return msg
		msg.highColorDepth = HIGH_COLOR_DEPTH(int.from_bytes(buff.read(2), byteorder='little', signed = False))
		if is_end(buff,start, dlen):
			return msg
		msg.supportedColorDepths = SUPPORTED_COLOR_DEPTH(int.from_bytes(buff.read(2), byteorder='little', signed = False))
		if is_end(buff,start, dlen):
			return msg
		msg.earlyCapabilityFlags = RNS_UD_CS(int.from_bytes(buff.read(2), byteorder='little', signed = False))
		if is_end(buff,start, dlen):
			return msg
		msg.clientDigProductId = buff.read(64)
		if RNS_UD_CS.VALID_CONNECTION_TYPE in msg.earlyCapabilityFlags:
			msg.connectionType = CONNECTION_TYPE(int.from_bytes(buff.read(1), byteorder='little', signed = False))
		else:
			buff.read(1)
			msg.connectionType = CONNECTION_TYPE.UNK
		if is_end(buff,start, dlen):
			return msg
		msg.pad1octet = buff.read(1)
		if is_end(buff,start, dlen):
			return msg
		msg.serverSelectedProtocol = SUPP_PROTOCOLS(int.from_bytes(buff.read(4), byteorder='little', signed = False))
		if is_end(buff,start, dlen):
			return msg
		msg.desktopPhysicalWidth = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		if is_end(buff,start, dlen):
			return msg
		msg.desktopPhysicalHeight = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		if is_end(buff,start, dlen):
			return msg
		msg.desktopOrientation = ORIENTATION(int.from_bytes(buff.read(2), byteorder='little', signed = False))
		if is_end(buff,start, dlen):
			return msg
		msg.desktopScaleFactor = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		if is_end(buff,start, dlen):
			return msg
		msg.deviceScaleFactor = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		return msg

	def __repr__(self):
		t = '==== TS_UD_CS_CORE ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t



import io
import enum
from aardwolf.protocol.T124.userdata.constants import TS_UD_TYPE, RNS_UD_SC

# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-rdpbcgr/379a020e-9925-4b4f-98f3-7d634e10b411
class TS_UD_SC_CORE:
	def __init__(self):
		self.type:TS_UD_TYPE = TS_UD_TYPE.SC_CORE
		self.length:int = None
		self.version:int = None
		self.clientRequestedProtocols:int = None
		self.earlyCapabilityFlags:RNS_UD_SC = None
		
	def to_bytes(self):
		def finish(t):
			t = (len(t)+4).to_bytes(2, byteorder='little', signed = False) + t
			t = self.type.value.to_bytes(2, byteorder='little', signed = False) + t
			return t
		t = self.version.to_bytes(4, byteorder='little', signed = False)
		t += self.clientRequestedProtocols.to_bytes(4, byteorder='little', signed = False)
		t += self.earlyCapabilityFlags.to_bytes(4, byteorder='little', signed = False)
		return finish(t)

	@staticmethod
	def from_bytes(bbuff: bytes):
		return TS_UD_SC_CORE.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		def is_end(buff, start, size):
			return buff.tell() - start >= size
		start = buff.tell()
		msg = TS_UD_SC_CORE()
		msg.type = TS_UD_TYPE(int.from_bytes(buff.read(2), byteorder='little', signed = False))
		msg.length = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		if is_end(buff,start, msg.length):
			return msg
		msg.version = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		if is_end(buff,start, msg.length):
			return msg
		msg.clientRequestedProtocols = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		if is_end(buff,start, msg.length):
			return msg
		msg.earlyCapabilityFlags = RNS_UD_SC(int.from_bytes(buff.read(4), byteorder='little', signed = False))
		if is_end(buff,start, msg.length):
			return msg
		return msg

	def __repr__(self):
		t = '==== TS_UD_SC_CORE ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t